class KurtosisParameters():
    """
    Kurtosis Parameters
    """
    def __init__(self, frequency_bands, window_lengths,
                 extrema_smoothings=[40], n_smooth=1):
        """
        :param frequency_bands: frequency bands to filter data before
            calculating kurtosis
        :type frequency_bands: list of [lo, high]s
        :param window_lengths:  window lengths (seconds) over which to
            calculate kurtosis
        :type window_lengths: list
        :param extrema_smoothings: kurtosis smoothing sequence (increasing
            list of samples)
        :type extrema_smoothings: list of ints
        :param n_smooth: samples over which to smooth kurtosis
        :type n_smooth: samples over which to smooth kurtosis
        """
        assert isinstance(frequency_bands, list)
        assert isinstance(window_lengths, list)
        assert isinstance(extrema_smoothings, list)
        for fb in frequency_bands:
            assert(len(fb) == 2)
        for x in window_lengths:
            assert isinstance(x, (int, float))
        for x in extrema_smoothings:
            assert isinstance(x, int)

        self.frequency_bands = frequency_bands
        self.window_lengths = window_lengths
        self.extrema_smoothings = extrema_smoothings
        self.n_smooth = n_smooth

    def __str__(self):
        str = "KurtosisParmeters("
        str += f"frequency_bands = {self.frequency_bands}, "
        str += f"window_lengths = {self.window_lengths}, "
        str += f"extrema_smoothings = {self.extrema_smoothings})"
        return str

    @classmethod
    def from_dict(cls, thedict):
        return cls(**thedict)



import warnings

from obspy import UTCDateTime


def write_sacpz(paz, net, sta, cha, fname, include_comments=False):
    """
    From obspy.io.sac.sacpz

    Does not modify normalization_factor, gives same output as converting
    stationxml to dataless using 'stationxml-converter -s', then dataless
    to SAC PZ using 'rdseed -p'

    :param paz: PoleZeroStage representing all PoleZeroStages in the response
    :sta: obspy Station object
    :cha: obspy Channel object
    :fname: filename to write to
    :include_comments: include SAC comments (ruins SEISAN read!)
    """
    sens = cha.response.instrument_sensitivity
    input_unit = sens.input_units.upper()
    if input_unit in ["M", "PA", "PASCALS"]:
        pass
    elif input_unit in ["M/S", "M/SEC"]:
        paz.zeros.append(0j)
    elif input_unit in ["M/S**2", "M/SEC**2"]:
        paz.zeros.extend([0j, 0j])
    else:
        msg = "{}.{}.{}.{} {} ".format(
            net.code,
            sta.code,
            cha.location_code,
            cha.code,
            cha.start_date)
        msg += "has unrecognized input units in "
        msg += "response: {}. Skipping".format(input_unit)
        warnings.warn(msg)
        return
    out = []
    if include_comments is True:
        out = get_comments(paz, net, sta, cha)
    out.append(_paz_to_sacpz_string(paz, sens))
    out.extend(["", ""])
    out = "\n".join(out) + "\n\n"
    with open(fname, "wt") as fh:
        fh.write(out)


def get_comments(paz, net, sta, cha):
    sens = cha.response.instrument_sensitivity
    input_unit = sens.input_units.upper()
    out = []
    now = UTCDateTime()
    out.append("* " + "*" * 50)
    out.append(f"* NETWORK     : {net.code}")
    out.append("* STATION     : %s" % sta.code)
    out.append("* LOCATION    : %s" % cha.location_code)
    out.append("* CHANNEL     : %s" % cha.code)
    out.append("* CREATED     : %s" % now.isoformat())
    out.append("* START       : %s" % cha.start_date.isoformat())
    out.append("* END         : %s" % cha.end_date.isoformat())
    out.append("* DESCRIPTION : %s" % sta.site.name)
    out.append("* LATITUDE    : %s" % (cha.latitude or sta.latitude))
    out.append("* LONGITUDE   : %s" % (cha.longitude or sta.longitude))
    out.append("* ELEVATION   : %s" % (cha.elevation or sta.elevation))
    out.append("* DEPTH       : %s" % cha.depth)
    # DIP in SACPZ served by IRIS SACPZ web service is
    # systematically different from the StationXML entries.
    # It is defined as an incidence angle (like done in SAC for
    # sensor orientation), rather than an actual dip.
    # Add '(SEED)' to clarify that we adhere to SEED convention
    out.append("* DIP (SEED)  : %s" % cha.dip)
    out.append("* AZIMUTH     : %s" % cha.azimuth)
    out.append("* SAMPLE RATE : %s" % cha.sample_rate)
    if input_unit in ['PA', 'PASCALS']:
        out.append(f"* INPUT UNIT  : {input_unit}")
    else:
        out.append("* INPUT UNIT  : M")
    out.append("* OUTPUT UNIT : %s" % sens.output_units)
    out.append("* INSTTYPE    : %s" % paz.description)
    #  out.append("* INSTTYPE    : %s" % cha.sensor.type)
    out.append(f"* INSTGAIN    : {paz.stage_gain:.6e} ({sens.input_units})")
    out.append(f"* SENSITIVITY : {sens.value:.6e} ({sens.input_units})")
    out.append("* A0          : {:.6e} ({})".format(paz.normalization_factor,
                                                    sens.input_units))
    out.append("* " + "*" * 50)
    return out


def _paz_to_sacpz_string(paz, sens):
    """
    Returns SACPZ ASCII text representation of Response.

    COPIED DIRECTLY FROM OBSPY.CORE.INVENTORY.RESPONSE

    :type paz: :class:`PolesZerosResponseStage`
    :param paz: Poles and Zeros response information
    :type instrument_sensitivity: :class:`InstrumentSensitivity`
    :param paz: Overall instrument sensitivity of response
    :rtype: str
    :returns: Textual SACPZ representation of poles and zeros response stage.
    """
    # assemble output string
    out = []
    out.append("ZEROS %i" % len(paz.zeros))
    for c in paz.zeros:
        out.append(" %+.6e %+.6e" % (c.real, c.imag))
    out.append("POLES %i" % len(paz.poles))
    for c in paz.poles:
        out.append(" %+.6e %+.6e" % (c.real, c.imag))
    out.append("CONSTANT {:.6e}".format(paz.normalization_factor * sens.value))
    return "\n".join(out)



import numpy as np
from scipy.signal import lfilter
# from ..logger import log

from obspy.core.stream import Stream
from obspy.core.stream import Trace


def smooth_filter(traces, n_smooth):
    """
    Smooth a trace or a list of traces

    Uses the scipy.signal.lfilter() function

    :param traces_in: trace of list of traces
    :param n_smooth: size of the smoothing window in SAMPLES
    :returns: smoothed data
    """
    n_smooth = int(n_smooth)
    bare_trace = False
    if isinstance(traces, Trace):
        traces = [traces]
        bare_trace = True
    assert isinstance(traces, list) or isinstance(traces, Stream)

    smoothed_traces = []
    for tr in traces:
        smooth_tr = tr.copy()
        smooth_tr.data = lfilter(np.divide(np.ones(n_smooth), n_smooth),
                                 1., tr.data)
        smoothed_traces.append(smooth_tr)
    if len(smoothed_traces) == 0:
        return None
    if bare_trace:
        return smoothed_traces[0]
    return smoothed_traces


if __name__ == '__main__':
    pass



import numpy as np
from scipy import signal, linalg
from obspy.core import UTCDateTime
from obspy.core.stream import Stream

from .logger import log
# from .timer import Timer


class Polarity:
    """
    polarity and dip-rect calculation
    """
    def __init__(self, stream, params, zcomponents='Z3', ncomponents='N1Y',
                 ecomponents='E2X', verbose=True):
        """
        :param stream: stream containing X, Y and Z traces
        :param params: PolarityParameters object
        :param zcomponents: possible characters for vertical components
        :param ncomponents: possible characters for "north" components
        :param ecomponents: possible characters for "east" components
        :param
        """
        stream, self.nsamps = self._clean_stream(stream)
        self._check_components(zcomponents, ncomponents, ecomponents)
        self.params = params
        self.verbose = verbose
        self.sr = stream[0].stats.sampling_rate
        self.tracez = None
        self.tracen = None
        self.tracee = None
        for tr in stream:
            if tr.stats.channel[-1] in zcomponents:
                assert self.tracez is None, 'already have a z trace!'
                self.tracez = tr.copy()
            elif tr.stats.channel[-1] in ncomponents:
                assert self.tracen is None, 'already have a n trace!'
                self.tracen = tr.copy()
            elif tr.stats.channel[-1] in ecomponents:
                assert self.tracee is None, 'already have a e trace!'
                self.tracee = tr.copy()
            else:
                msg = f"Unexpected channel code {tr.stats.channel}"
                raise ValueError(msg)

    @staticmethod
    def _clean_stream(st):
        """
        Validate stream and synchronize traces
        """
        assert len(st) == 3, 'stream does not have three channels'
        assert st[0].stats.sampling_rate == st[1].stats.sampling_rate,\
            'traces have different sampling rates'
        assert st[1].stats.sampling_rate == st[2].stats.sampling_rate,\
            'traces have different sampling rates'

        stime = UTCDateTime(np.max([x.stats.starttime.timestamp for x in st]))
        etime = UTCDateTime(np.min([x.stats.endtime.timestamp for x in st]))
        st.trim(stime, etime)
        n_samps = np.min([len(x.data) for x in st])
        for tr in st:
            if not len(tr.data) == n_samps:
                tr.data = tr.data[:n_samps]
        return st, n_samps

    @staticmethod
    def _check_components(zc, nc, ec):
        """
        Verify that there are no repeated characters in the component maps

        :param zc: characters allowed for z component
        :param nc: characters allowed for n component
        :param ec: characters allowed for e component
        """
        error_str = "overlap between {} and {} component maps: {}, {}"
        for c in zc:
            assert c not in nc, error_str.format('z', 'n', zc, nc)
            assert c not in ec, error_str.format('z', 'e', zc, ec)
        for c in ec:
            assert c not in nc, error_str.format('e', 'n', ec, nc)

    def calc_dip_rect(self, times, min_dip_threshold=30, plot=False):
        """
        Return dip-rect values after specified times

        Dip-rect is the rectilinearity multiplied by -1 for near-horizontal
            particle motions, by 1 for near-vertical particle motions.

        :param times: list of UTCDateTimes
        :param min_dip_threshold: do not calculate dip-rectilinearity if there
            is no calculated dip with at least this absolute angle (degrees)
        :returns: DR: Dip-rectilinearity trace
        """
        if len(times) == 0:
            return None
        # Pick_Function.m:543
        # with Timer(text="    polarity.calc_dip_rect(): {:0.4f}s"):
        rectP, aziP, dipP = self.polar_analysis(times)
        if np.max(np.abs(dipP.data)) < min_dip_threshold:
            if self.verbose:
                log("max dip ({:.1f}) < threshold ({:g}), no DR calculated".
                    format(np.max(np.abs(dipP.data)), min_dip_threshold),
                    "verbose")
            return None

        dipp = np.sin(np.abs(np.radians(dipP.data)))  # 1 for vert, 0 for hor
        smooth_samples = int(np.round(self.params.DR_smooth_length * self.sr))
        if smooth_samples < 2:
            smooth_samples = 2
        a = np.ones(smooth_samples) / smooth_samples
        a2 = np.ones(2*smooth_samples) / (2 * smooth_samples)
        smooth_dipp = signal.lfilter(a, 1, dipp)
        smooth_rectP = signal.lfilter(a, 1, rectP.data)
        Drb = np.sign(1.3 * smooth_dipp - smooth_rectP)  # pos for P, neg for S
        DR = rectP.copy()
        DR.data *= Drb
        DR.data = signal.lfilter(a2, 1, DR.data)
        # DR.data = signal.lfilter(a2, 1, np.multiply(rectP.data, Drb))
        DR.data[rectP.data == 0] = 0
        if plot:
            DR.stats.channel = 'DR'
            dipP.stats.channel = 'DIP'
            Stream([DR, dipP, self.tracez, self.tracen, self.tracee]).plot(
                equal_scale=False)
        return DR

    def polar_analysis(self, times):
        """
        Compute the polarity parameters around specified times

        Only computes around user specified times since this analysis is time
        consuming.  Other values are set to zero
        :param times: list of times around which to compute polarity
        :returns: rect, azi (degrees), dip (degrees)
        """
        # fast_polar_analysis.m:17
        # with Timer(text="polarity.polar_analysis() calc_indices: {:0.4f}s"):
        ind_vec, n_half_analyze = self._calc_indices(times)

        # with Timer(text="    polarity.polar_analysis() rest: {:0.4f}s"):
        rectP = self.tracez.copy()
        rectP.data[:] = 0.
        aziP = rectP.copy()
        dipP = rectP.copy()
        for k in ind_vec:
            eP = self.tracee.data[k - n_half_analyze: k + n_half_analyze]
            nP = self.tracen.data[k - n_half_analyze: k + n_half_analyze]
            zP = self.tracez.data[k - n_half_analyze: k + n_half_analyze]
            MP = np.cov(np.array([eP, nP, zP]))
            D, V = linalg.eig(MP)
            i_sort = np.argsort(D)  # sort from min (0) to max (2)
            D = np.abs(D[i_sort])
            V = V[:, i_sort]
            # 2 = major axis, 0 = minor axis, 1 = intermediate
            rectP.data[k] = 1 - ((D[0] + D[1]) / (2 * D[2]))
            V_major = V[:, 2]
            aziP.data[k] = np.degrees(np.arctan(V_major[1] / V_major[0]))
            dipP.data[k] = np.degrees(np.arctan(V_major[2]
                                                / np.sqrt(V_major[1]**2
                                                          + V_major[0]**2)))
        return rectP, aziP, dipP

    def _calc_indices(self, pick_times):
        """
        Create list of indices to investigate

        Compute windows after picks minus duplicate samples or samples too
        close to the data edges
        """
        n_compute = round(self.sr * self.params.calculate_window)
        ind_vec = np.array([], dtype='int32')
        # ztimes = self.tracez.times('utcdatetime')
        ztimes = self.tracez.times('timestamp')
        for t in pick_times:
            index = ztimes.searchsorted(t.timestamp)
            iwind = np.arange(index, index + n_compute, dtype='int32')
            ind_vec = np.concatenate([ind_vec, iwind])

        n_half_analyze = int(round(self.params.analyze_window * self.sr / 2))
        # Get rid of duplicate or out-of-bounds values
        ind_vec = ind_vec[ind_vec - n_half_analyze >= 0]
        ind_vec = ind_vec[ind_vec + n_half_analyze < self.nsamps]
        ind_vec = np.sort(ind_vec)

        # ind_vec[np.diff(ind_vec, prepend=-1) == 0] = []
        ind_vec = ind_vec[np.diff(ind_vec, prepend=-1) != 0]
        # print(ind_vec, n_half_analyze)
        return ind_vec, n_half_analyze

    if __name__ == '__main__':
        pass



# import warnings
import itertools

import numpy as np
from scipy.cluster.hierarchy import fclusterdata
from obspy.taup import TauPyModel
from obspy.taup.taup_create import build_taup_model
from obspy.taup.tau_model import TauModel
from obspy.core import UTCDateTime
# from obspy.core.event.origin import Pick

from .utils import picks_matched_stations
from .logger import log
from .pick_candidate import PickCandidate


class Associator():
    """
    Associate Picks

    Currently only rejects picks that deviate from expected distribution
    ToDo:
        - Replace rejected picks with other candidates
        - Add velocity model-based selection criteria
    """
    # Values from Kennett 1995 for 0, 1, 5, 10, 30 and 90 degrees
    # default_delays = {'op': [0, 19.17, 76.27, 144.90, 370.27, 781.35],
    #                   'ps': [0, 13.92, 59.63, 114.2, 300.00, 654.43]}

    def __init__(self, params, vel_model=None, target_depth=10., max_offset=1.,
                 verbose=True, debug=False):
        """
        :param params: AssociatorParameters object
        :param vel_model: velocity model (obspy taup valid string or the
            absolute path name of a .nd file)
        :param target_depth: event depth to use to construct travel-time tables
        :param max_offset: maximum offset(degrees) for which to calculate
            travel_times

        The nd file format allows c-style comment lines, depth-parameter lines
        and named discontinuity lines.
        The depth-parameter lines have space separated columns for:
        Depth(km), Pvel(km/s), Svel(km/s) and, optional density(g/cc), Qp & Qs
        The named discontinuity lines are placed between the depth-parameter
        lines corresponding to the values above and below.  For example:
        // This is a comment
        0.0    5.0   3.0
        20     5.0   3.0
        20     6.5   3.7
        33     6.5   3.7
        mantle        # the word "mantle" designates that this is the moho
        33     7.8   4.4
        410    8.9   4.7
        """
        self.method = params.method
        self.distri_min_values = params.distri_min_values
        self.cluster_window_P = params.cluster_window_P
        self.cluster_window_S = params.cluster_window_S
        self.distri_nstd_picks = params.distri_nstd_picks
        self.distri_nstd_delays = params.distri_nstd_delays
        self.cluster_window_otime = params.cluster_window_otime
        self.vp_over_vs = params.otime_vp_vs
        self.p_cluster = {}   # a dictionary with key=station and value=ptime
        self.s_cluster = {}   # a dictionary with key=station and value=stime
        self.ps_cluster = {}   # a dictionary with key=station and value=stime
        self.o_cluster = {}   # a dictionary with key=station and value=otime
        self.delays = None
        if vel_model is not None:
            try:
                self.delays = self._get_delays(vel_model, target_depth,
                                               max_offset)
            except Exception:
                log("Couldn't generate ps delays using '" + vel_model
                    + "', using default", "warning")
        self.verbose = verbose
        self.debug = debug
        self.min_clust = 3  # minimum number of values in a "cluster"

    def run(self, picks, candidates):
        """
        Run associator

        Tries to associate by origin time, if that doesn't work associates by
        pick clustering
        :param picks: list of preferred picks
        :param candidates: all pick candidates (including preferred picks)

        Future upgrade: associate by position (requires velocity model and
        station positions)
        """
        picks = PickCandidate.remove_duplicates(picks)
        input_picks = picks.copy()
        method = self.method
        assert method in ['origin_time', 'arrival_time']
        if method == 'origin_time':
            assoc_str = 'origin times'
            picks, associated = self.find_same_origin_time(picks, candidates)
            if not associated:
                log('Could not associate by origin time', 'verbose')
                method = 'pick_time'
        if method == 'arrival_time':
            assoc_str = 'pick clustering'
            picks = self.remove_nonclustered(picks)
        picks = PickCandidate.remove_duplicates(picks)
        self._assoc_stats(assoc_str, input_picks, picks)
        return picks

    @staticmethod
    def _assoc_stats(assoc_type, input_picks, output_picks):
        """
        Print information about accepted, rejected, added and changed picks
        """
        ops = output_picks.copy()
        stats = {'accepted': [], 'rejected': [], 'modified': [], 'added': []}
        # n_changes = 0
        for p in input_picks:
            if p in ops:
                stats['accepted'].append(p.shortname)
                ops.remove(p)
            else:
                op_remove = None
                for op in ops:
                    if (p.station == op.station
                            and p.phase_guess == op.phase_guess):
                        stats['modified'].append(p.shortname)
                        op_remove = op
                if op_remove is not None:
                    ops.remove(op_remove)
                else:
                    stats['rejected'].append(p.shortname)
        for p in ops:
            stats['added'].append(p.shortname)

        s = f'Associated by {assoc_type}: '
        for key in stats:
            s += f'{len(stats[key]):d} {key}, '
        log(s, 'verbose')
        for key in ['rejected', 'modified', 'added']:
            if len(stats[key]) > 1:
                log(f"{key.upper()}: {stats[key]}", 'debug')

    def remove_nonclustered(self, picks):
        """
        Very basic (clustering-based) pick selection by association

        :param picks: list of obspy Picks
        :candidates: dictionary with key=station, value=list of PickCandidates
        """
        # Remove lines with arrival times = start_time
        # P_picked_cell=rm_cell_line(P_picked_cell[2] == 0,P_picked_cell)
        # S_picked_cell=rm_cell_line(S_picked_cell[2] == 0,S_picked_cell)

        picks = self._remove_unclustered(picks)
        stations = _pick_stations(picks)
        if self.distri_min_values <= len(stations):
            picks = self._remove_badly_distributed(picks)
            picks = self._remove_bad_delays(picks)
        return picks

    def find_same_origin_time(self, picks, candidates):
        """
        Select picks by origin time

        :param picks: list of PickCandidates pre-selected as P and S picks
        :candidates: list of all PickCandidates

        :returns: list of picks, whether associator worked
        :rtype: list of PickCandidate, bool
        """
        stations = _pick_stations(picks)
        ots = []
        for station in stations:
            station_picks = [x for x in picks
                             if x.station == station]
            p_pick = [x for x in station_picks if x.phase_guess == "P"]
            s_pick = [x for x in station_picks if x.phase_guess == "S"]
            if not (len(p_pick) == 1 and len(s_pick) == 1):
                continue
            ots.append(self._ps_to_otime(p_pick[0].time, s_pick[0].time))
        if len(ots) < self.min_clust:
            log(f'less than {self.min_clust} P-S calculated origin times'
                ', cannot associate by this criteria', 'verbose')
            for p in picks:
                log(f' {p}', 'verbose')
            return picks, False
        good_ots = self._cluster_clean_otimes([t for t in ots])
        if len(good_ots) < self.min_clust:
            log('less than {} P-S origin times agree, cannot associate'
                .format(self.min_clust), 'verbose')
            for p in picks:
                log(f' {p}', 'debug')
            return picks, False
        mean_ot = UTCDateTime(np.mean([x.timestamp for x in good_ots]))
        new_picks = self._find_otime_matching(mean_ot, picks, candidates)
        return new_picks, True

    def _find_otime_matching(self, ot, picks, candidates):
        """
        Return picks or candidates whose P-S delay matches the origin time

        :param ot: desired origin time
        :param picks: list of preferred PickCandidates
        :param candidates: list of all PickCandidates
        :returns: new list of preferred PickCandidates
        """
        otime_margin = self.cluster_window_otime / 2.
        stations_c = _pick_stations(candidates)
        # Error check
        for sta in _pick_stations(picks):
            if sta not in stations_c:
                raise ValueError(f"picked station {sta} not in candidates")

        new_picks = []
        self.o_cluster = {}
        for station in stations_c:
            sta_candidates = [x for x in candidates if x.station == station]
            p_existing, s_existing = self._get_existing(picks, station)

            # If we have a P and an S candidate that match ot, keep them
            if s_existing is not None and p_existing is not None:
                # log(f'{station}: s_existing and p_existing', 'debug')
                otime = self._ps_to_otime(p_existing.time, s_existing.time)
                if np.abs(otime - ot) < otime_margin:
                    new_picks.extend([p_existing, s_existing])
                    self.o_cluster[station] = otime
                    continue
            # Otherwise, if a P cand matches another cand's ot, keep them
            if p_existing is not None:
                # log(f'{station}: p_existing top', 'debug')
                # log(p_existing, 'debug')
                cands = [x for x in sta_candidates if not x == p_existing]
                if len(cands) > 0:
                    otimes = [self._ps_to_otime(p_existing.time, x.time)
                              for x in cands]
                    offsets = [np.abs(x - ot) for x in otimes]
                    if np.min(offsets) < otime_margin:
                        s_candidate = cands[np.argmin(offsets)]
                        s_candidate.phase_guess = 'S'
                        # log(p_existing, 'debug')
                        # log(s_candidate, 'debug')
                        new_picks.extend([p_existing, s_candidate])
                        self.o_cluster[station] = otimes[np.argmin(offsets)]
                        continue
            # Otherwise, an S cand that matches with another cand, keep them
            if s_existing is not None:
                # log(f'{station}: s_existing top', 'debug')
                cands = [x for x in sta_candidates if not x == s_existing]
                if len(cands) > 0:
                    otimes = [self._ps_to_otime(x.time, s_existing.time)
                              for x in cands]
                    offsets = [np.abs(x - ot) for x in otimes]
                    if np.min(offsets) < otime_margin:
                        p_candidate = cands[np.argmin(offsets)]
                        p_candidate.phase_guess = 'P'
                        new_picks.extend([p_candidate, s_existing])
                        self.o_cluster[station] = otimes[np.argmin(offsets)]
                        continue
            # Otherwise, if any combination of cands matches ot, keep them
            if sta_candidates is not None:
                # log(f'{station}: neither s_existing nor p_existing', 'debug')
                sort_cands = sorted(sta_candidates, key=lambda x: x.time)
                min_offset = otime_margin
                new_p, new_s = None, None
                for p_maybe, s_maybe in itertools.combinations(sort_cands, 2):
                    otime = self._ps_to_otime(p_maybe.time, s_maybe.time)
                    offset = np.abs(otime - ot)
                    if offset < min_offset:
                        new_p = p_maybe
                        new_s = s_maybe
                        new_otime = otime
                        min_offset = offset
                if new_p is not None:
                    new_p.phase_guess = 'P'
                    new_s.phase_guess = 'S'
                    new_picks.extend([new_p, new_s])
                    self.o_cluster[station] = new_otime
                    continue
            # Otherwise, keep a solitary P or S pick
            # (if I used station positions, I could compare times here)
            if p_existing is not None and s_existing is None:
                # log(f'{station}: p_existing bottom', 'debug')
                assert isinstance(p_existing, PickCandidate)
                new_picks.append(p_existing)
            elif p_existing is None and s_existing is not None:
                # log(f'{station}: s_existing bottom', 'debug')
                assert isinstance(s_existing, PickCandidate)
                new_picks.append(s_existing)
        # log('about to test for duplicates', 'debug')
        # new_picks = PickCandidate.remove_duplicates(new_picks)
        # log('just tested for duplicates', 'debug')
        return new_picks

    @staticmethod
    def _get_existing(picks, station):
        p_existings = [x for x in picks if x.station == station
                       and x.phase_guess == 'P']
        s_existings = [x for x in picks if x.station == station
                       and x.phase_guess == 'S']
        p_existing, s_existing = None, None
        if len(p_existings) > 0:
            p_existing = p_existings[0]
        if len(s_existings) > 0:
            s_existing = s_existings[0]
        return p_existing, s_existing

    def _remove_unclustered(self, picks):
        """
        Remove picks and reassign phases based on clustering

        Very basic association
        :param picks: list of PickCandidate objects
        :returns: modified list
        """
        # Pick_Function.m:746
        p_picks = [p for p in picks if p.phase_guess == 'P']
        if len(p_picks) >= self.distri_min_values:
            p_picks = self._cluster_clean_picks('P'. p_picks)
            log(f'clustered p_picks = {p_picks}', 'debug')
            self.p_cluster = {x.station: x.time for x in p_picks}

        s_picks = [p for p in picks if p.phase_guess == 'S']
        if len(s_picks) >= self.distri_min_values:
            s_picks = self._cluster_clean_picks('S', s_picks)
            log(f'clustered s_picks = {s_picks}', 'debug')
            self.p_cluster = {x.station: x.time for x in s_picks}

        return p_picks + s_picks

    def _remove_badly_distributed(self, picks):
        """
        Remove picks that are well outside of pick distribution

        :param picks: input Picks
        """
        p_picks = [p for p in picks if p.phase_guess == 'P']
        s_picks = [p for p in picks if p.phase_guess == 'S']
        _, iP = clean_distri([x.time.timestamp for x in p_picks],
                             self.distri_nstd_picks, 'median',
                             self.distri_min_values)
        _, iS = clean_distri([x.time.timestamp for x in s_picks],
                             self.distri_nstd_picks, 'median',
                             self.distri_min_values)
        return [p_picks[i] for i in iP] + [s_picks[i] for i in iS]

    def _remove_bad_delays(self, picks):
        """
        Eliminate picks based on distribution of P-S delays
        """
        # Pick_Function.m:763
        matches = picks_matched_stations(picks)
        if len(matches) == 0:
            return picks
        delay_stations = [m['station'] for m in matches]
        delays = [m['pickS'].time - m['pickP'].time for m in matches]
        _, i_PS = clean_distri(delays, self.distri_nstd_delays,
                               'median', self.distri_min_values)
        # Include good delay stations AND non-delay stations
        good_delay_stations = [delay_stations[i] for i in i_PS]
        good_delays = [delays[i] for i in i_PS]
        self.ps_cluster = {s: x for s, x in zip(good_delay_stations,
                                                good_delays)}

        bad_delay = [s for s in delay_stations if s not in good_delay_stations]
        good_picks = [p for p in picks if p.station not in bad_delay]
        return good_picks

    def _get_origtime_delays(self, vel_mod, target_depth, max_dist=1.,
                             num_vals=20):
        """
        :param max_dist: maximum distance (degrees) at which to calculate
        :param num_vals: number of distances (from 0 to max_distance) at
            which to calculate delats
        :returns: dict with keys 'op' and 'ps' corresponding to sorted lists
            of delay times from origin-to-p and p-to-s
        """
        try:
            model = TauPyModel(model=vel_mod)
        except Exception:
            build_taup_model(vel_mod)   # converts from np or tvel to npz
            model = TauModel.from_file(vel_mod)  # reads npz
        delays = {'op': [0], 'ps': [0]}
        # ps_delays, p_delays = [0], [0]
        step = max_dist / num_vals
        for distance in np.arange(step, max_dist + .001, step):
            arrivals = model.get_travel_times(target_depth, distance,
                                              phase_list=["P", "S"])
            p_delay = [x.time for x in arrivals if x.phase == "P"]
            s_delay = [x.time for x in arrivals if x.phase == "S"]
            ps_delay = s_delay - p_delay
            if len(ps_delay) == 1 and len(p_delay) == 1:
                delays['ps'].append(ps_delay[0])
                delays['op'].append(p_delay[0])
            else:
                log('Could not calculate delays for dist={:.2g} degrees'
                    .format(distance), 'warning')
        return delays

    def _ps_to_otime(self, p_time, s_time):
        """
        Calculate origin time given P and S arrival times

        :param p_time: p arrival time
        :param s_time: s arrival time
        """
        return self.calc_origin_time(p_time, s_time, self.vp_over_vs,
                                     self.delays)

    @staticmethod
    def calc_origin_time(p_time, s_time, vpvs=1.65, delays_model=None):
        """
        Calculate origin time given P and S arrival times

        :param p_time: P arrival time (UTCDateTime)
        :param s_time: S arrival time (UTCDateTime)
        :param delays_model: dict with keys='op' and 'ps', each with an
            equal-length sorted list of origin-to-P and P-to-S delays
        :param vpvs: Vp/Vs ratio, to be used if delays_model == None, using the
            equation  o_time = p_time - ps_delay / (vp/vs - 1)
        :returns: origin_time
        :rtype: UTCDateTime
        """
        origin_time = p_time
        if delays_model is not None:
            origin_time -= np.interp(s_time-p_time,
                                     delays_model['ps'],
                                     delays_model['op'])
        else:
            origin_time -= (s_time - p_time) / (vpvs - 1)
        return origin_time

    def _cluster_clean_otimes(self, times=None):
        """
        Return indices of origin times fitting in largest cluster
        """
        indices = cluster_clean_indices(self.min_clust,
                                        self.cluster_window_otime,  times)
        return [times[i] for i in indices]

    def _cluster_clean_picks(self, phase=None, picks=None):
        """
        Return picks fitting in the largest cluster

        :param phase: "P" or "S"
        :param picks: list of obspy Picks
        :returns: picks that fit the cluster criteria
        """
        assert phase in 'PS', f'phase is "{phase}", should be "P" or "S"'
        if phase == 'P':
            window_sec = self.cluster_window_P
        else:
            window_sec = self.cluster_window_S
        indices = cluster_clean_indices(self.min_clust, window_sec,
                                        [x.time for x in picks])
        return [picks[i] for i in indices]


def cluster_clean_indices(min_clust=None, window_sec=None, times=None):
    """
    Return indices of times fitting in the largest cluster

    :param min_clust: minimum cluster size
    :param window_sec: cluster window length
    :param times: list of UTCDateTime

    :returns: indices of times that fit the cluster criteria
    """
    if len(times) == 0:
        return []
    elif len(times) == 1:
        return [0]
    else:
        cluster_data = fclusterdata(np.array([[x.timestamp for x in times]]).T,
                                    t=window_sec, criterion='distance')
        cluster_groups = dict()
        for cnum in cluster_data:
            cluster_groups[cnum] = cluster_groups.get(cnum, 0) + 1
        max_length = max(cluster_groups.values())
        max_keys = [k for k, v in cluster_groups.items() if v == max_length]

        if len(max_keys) > 1 and max_length >= min_clust:
            log('{:d} clusters of max_length ({:d}), returning first one'
                .format(len(max_keys), max_length), 'warning')
            log(np.nonzero(cluster_data == max_keys[0])[0], 'debug')
        indices = np.nonzero(cluster_data == max_keys[0])[0]
        return indices


def clean_distri(values, n_std=3, mode='median', min_vals=3):
    """
    Remove extreme values in a vector

    :param values: list of values
    :param n_std: erase values for which x is more than n_std standard
        deviations from mode(x)
    :param mode: 'mean' or 'median' [ default = 'median' ]
    :param min_vals:  do not clean if vector is shorter than this[default=3]
    :returns: list of values after cleaning, indices of values retained

    >>> clean_distri([0, 3, 4, 6, 3, 5, 25])
    (array([0, 3, 4, 6, 3, 5]), array([0, 1, 2, 3, 4, 5]))
    """
    if len(values) < min_vals:
        return values, list(range(len(values)))

    x = np.array(values)
    if mode == 'mean':
        dev = x - np.mean(x)
    else:
        dev = x - np.median(x)

    # Calculate std of values
    lim = round(0.68*len(x))  # assume gaussian around median
    if lim == 0:
        lim = 1
    std = np.sort(np.abs(dev))[lim]

    in_rm = np.nonzero(np.abs(dev) <= n_std * std)[0]

    return x[in_rm], in_rm


def _pick_stations(picks):
    # log(picks, 'debug')
    return list(set([x.station for x in picks]))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# Generated with SMOP  0.41-beta
# from smop.libsmop import *
# readmain.m


class SNRParameters():
    """
    Picker Parameters
    """
    def __init__(self,
                 signal_window,
                 noise_window,
                 quality_thresholds,
                 max_threshold_crossings=5,
                 threshold_parameter=0.2):
        """
        Initialize signal-to-noise ratio parameters

        :param noise_window: Length of noise window in seconds
        :param signal_window: Length of signal window in seconds
        :param quality_thresholds: SNR thresholds for pick quality
            list with values for quality = '3', '2', '1' and '0', in order
            the minimum value is also used as a baseline for SNR_threshold
        :param threshold_parameter: Threshold for SNR-based quality
            evaluation.
            if > 0 and less than 1, then SNR_threshold is this times the
                maximum SNR
            if < 0, then SNR_threshold is abs(SNR_threshold_parameter)
        :param max_threshold_crossings: Maximum crossings of SNR threshold
            level to accept a trace
        """
        qt = quality_thresholds
        assert len(qt) == 4, f'Need 4 quality_thresholds, found {len(qt):d}'
        assert qt == sorted(qt), f"quality thresholds not increasing: {qt}"

        self.signal_window = signal_window
        self.noise_window = noise_window
        self.quality_thresholds = quality_thresholds
        self.max_threshold_crossings = max_threshold_crossings
        self.threshold_parameter = threshold_parameter

    def __str__(self):
        str = "SNRParameters:\n"
        str += f"    signal_window = {self.signal_window}\n"
        str += f"    noise_window = {self.noise_window}\n"
        str += f"    quality_thresholds = {self.quality_thresholds}\n"
        str += "    max_threshold_crossings = "
        str += f"{self.max_threshold_crossings}\n"
        str += f"    threshold_parameter = {self.threshold_parameter}\n"
        return str

    @classmethod
    def from_dict(cls, thedict):
        return cls(**thedict)


if __name__ == '__main__':
    pass



"""
Routines involving lists of obspy Picks
"""


def picks_ps_times(picks):
    """
    Return ps_times and p_time for events having P and S picks

    :param picks: list of obspy Pick objects
    :returns: delays, P_times
    :rtype: list, list of UTCDateTime
    """
    matches = picks_matched_stations(picks)
    if len(matches) == 0:
        return None, None
    return ([(x['pickS'].time - x['pickP'].time) for x in matches],
            [x['pickP'].time for x in matches])


def picks_matched_stations(picks):
    """
    Return list of stations with P AND S picks

    :param picks: list of obspy Pick objects
    :returns: list of {'station', 'pickP', 'pickS'}
    """
    p_picks = [p for p in picks if p.phase_hint[0] == 'P']
    s_picks = [p for p in picks if p.phase_hint[0] == 'S']
    matches = []
    for p in p_picks:
        for s in s_picks:
            if p.waveform_id.station_code == s.waveform_id.station_code:
                matches.append({'station': p.waveform_id.station_code,
                                'pickP': p, 'pickS': s})
                break
    return matches

# def _old_match_pick_stations(picksP, picksS):
#     """
#     Return list of stations with P AND S picks, and the indices of the picks
#
#     :param picksP: list of P picks
#     :param picksS: list of S picks
#     :returns: list of {'station', 'iP', 'iS'}
#     """
#     matches = []
#     iP = 0
#     for pickP in picksP:
#         iS = 0
#         for pickS in picksS:
#             if pickP.station == pickS.station:
#                 matches.append({'station': pickP.station,
#                                 'iP': iP, 'iS': iS})
#                 break
#             iS += 1
#         iP += 1
#     return matches



"""
These routines should never need to be called by the user
"""
from .pick_utils import picks_matched_stations, picks_ps_times
from .select_traces import select_traces
from .smooth_filter import smooth_filter

__all__ = ['select_traces', 'smooth_filter',
           'picks_matched_stations', 'picks_ps_times']



# from readMSEEDTraces.m  ?
import re

from ..channel_map import ChannelMap
from ..logger import log


def select_traces(stream, map_rules, debug=False):
    """
    Map traces and phases

    selects traces corresponding to "Z", "N", "E" and "H", names the
    components to write picks to and defines the phase names
    :param stream: obspy stream
    :param map_rules: ChannelMappingRules object
    :param verbose: talk
    :returns: channel_mappings: sorted dict of ChannelMap with key=station
    """
    channel_map = dict()
    stations = list(set([t.stats.station for t in stream]))
    mr = map_rules
    for station in sorted(stations):
        Z = findChannel(stream, station, mr.compZ, mr.band_order)
        # N = findChannel(stream, station, mr.compN, mr.band_order)
        # E = findChannel(stream, station, mr.compE, mr.band_order)
        # H = findChannel(stream, station, mr.compH, mr.band_order)
        if Z is None:
            sta_st = stream.select(station=station)
            log('No Z channel found for station {}: {} not in "{}"'.format(
                station, [t.stats.channel[-1] for t in sta_st], mr.compZ),
                'error')
        else:
            channel_map[station] = ChannelMap(
                Z=findChannel(stream, station, mr.compZ, mr.band_order),
                N=findChannel(stream, station, mr.compN, mr.band_order),
                E=findChannel(stream, station, mr.compE, mr.band_order),
                H=findChannel(stream, station, mr.compH, mr.band_order),
                P_write_cmp=mr.P_write_cmp,
                S_write_cmp=mr.S_write_cmp,
                P_write_phase=mr.P_write_phase,
                S_write_phase=mr.S_write_phase)

    if debug:
        print('Channel Map:')
        print(f'{"Station":8s}|'
              + channel_map[stations[0]].__str__(format='table_header'))
        for station in sorted(stations):
            print(f'{station:8s}|'
                  + channel_map[station].__str__(format='table_row'))
    return channel_map


def findChannel(stream, station, cmp_chars, band_order):
    """
    Return the seed ID matching the given station and channel regex

    :param station: station name
    :param cmp_chars: possible characters for component code
    :band order: string listing band codes in order of preference (used only
        if there is more than one trace with valid component code)
    :returns: None if none found, error if more than one found
    """
    id_match = [tr.get_id() for tr in stream.select(station=station)
                if re.search(f'[{cmp_chars}]', tr.stats.channel[-1])
                is not None]
    if len(id_match) > 1:
        for band_code in band_order:
            if band_code in [id.split('.')[-1][0] for id in id_match]:
                id_match = [id for id in id_match
                            if id.split['.'][0] == band_code]
    if len(id_match) == 0:
        return None
    assert len(id_match) == 1,\
        'more than one match found for station = {}, cmp = {}'.format(
            station, cmp_chars)
    return id_match[0]


if __name__ == '__main__':
    pass



# Generated with SMOP  0.41-beta
# from smop.libsmop import *
# readmain.m


class AssociatorParameters():
    """
    Associator Parameters
    """
    def __init__(self,
                 cluster_window_otime,
                 cluster_window_P,
                 cluster_window_S,
                 method='origin_time',
                 otime_vp_vs=1.75,
                 distri_min_values=4,
                 distri_nstd_picks=3.2,
                 distri_nstd_delays=4):
        """
        Initialize Associator Parameters

        :param cluster_window_otime: Window length in seconds for cluster-based
            rejection of origin times
        :param otime_vp_vs: Vp/Vs value to use for origin time calculations
        :param cluster_window_P: Window [s] for P-pick rejection
                                       based on clustering
        :param cluster_window_S: Same as above, for S-pick
        :param distri_min_values: Minimum number of values (P-picks,
            S-picks or PS delays) needed to perform distribution-based
            rejection (if > n_stations of n_picks, will not perform
            distribution-based rejection
        :param distri_nstd_picks: maximum number of deviations from
            standard distribution to accept for P picks, and for S picks
        :param distri_nstd_delays: same as above, for P-S delays
        """
        assert method in ['origin_time', 'arrival_time']
        self.method = method
        self.cluster_window_otime = otime_vp_vs
        self.otime_vp_vs = otime_vp_vs
        self.cluster_window_P = cluster_window_P
        self.cluster_window_S = cluster_window_S
        self.distri_min_values = float(distri_min_values)
        self.distri_nstd_picks = distri_nstd_picks
        self.distri_nstd_delays = distri_nstd_delays

    def __str__(self):
        str = "AssociatorParameters:\n"
        str += f"    method = '{self.method}\n"
        str += f"    cluster_window_orig = {self.cluster_window_otime}\n"
        str += f"    otime_vp_vs = {self.otime_vp_vs}\n"
        str += f"    cluster_window_P = {self.cluster_window_P}\n"
        str += f"    cluster_window_P = {self.cluster_window_P}\n"
        str += f"    cluster_window_S = {self.cluster_window_S}\n"
        str += f"    distri_min_values = {self.distri_min_values}\n"
        str += f"    distri_nstd_picks = {self.distri_nstd_picks}\n"
        str += f"    distri_nstd_delays = {self.distri_nstd_delays}\n"
        return str

    @classmethod
    def from_dict(cls, thedict):
        return cls(**thedict)


if __name__ == '__main__':
    pass



from .kurtosis_parameters import KurtosisParameters


class StationParameters():
    """
    Station Parameters
    """
    def __init__(self, P_comp, S_comp, energy_frequency_band, energy_window,
                 kurtosis, resp_file, use_polarity=False, lag=999,
                 max_candidates=5):
        """
        :param P_Comp: String of components used to search for P arrival
            & amplitude
        :param S_comp: String of components used to search for S arrival
            & amplitude
        :param energy_frequency_band: frequency band in which to calculate
            SNR [lo, high]
        :param energy_window: only look at data from t-winlen to t, where t
            is the time of the peak waveform energy.  If == 0, don't use
            energy criteria.
        :param kurtosis: KurtosisParameters dictionary
        :param resp_file: response file name
        :param lag: unused
        :param use_polarity: Use polarity analysis to identify phases?
        :param max_candidates: maximum number of extrema to pick as candidates
        """
        self.P_comp = P_comp
        self.S_comp = S_comp
        self.energy_frequency_band = energy_frequency_band
        self.kurtosis = KurtosisParameters(**kurtosis)
        self.energy_window = energy_window
        self.resp_file = resp_file
        self.use_polarity = use_polarity
        self.lag = lag
        self.max_candidates = max_candidates

    @classmethod
    def from_dict(cls, thedict):
        return cls(**thedict)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from obspy.core.stream import Stream

from .logger import log
from .associator import Associator


class Plotter():
    """
    Class to plot figures representing the work in PSPicker
    """
    def __init__(self, plot_global=True, plot_stations=True):
        """
        :param plot_global: show global selection and picks plots
        :kind plot_global: bool
        :param plot_stations: show individual station plots
        :kind plot_stations: bool
        """
        self.gw = Global_Window(plot_global)
        self.pw = Picks_Window(plot_global)
        self.sw = Station_Window(plot_stations)


class Global_Window():
    """
    Global window Plotter

    Shows the entire time span for all stations, the initial picks and the
    selected pick window
    """
    def __init__(self, plot=True):
        self.plot = plot
        if not plot:
            self.fig, self.ax = None, None
        else:
            self.fig, self.ax = plt.subplots(num="PSPicker: Overview",
                                             clear=True)
        self.stations = []  # ordered list of stations

    def setup(self, starttime, endtime, stations):
        """
        :param starttime: plot start time
        :param endtime: plot end time
        :param stations: sorted list of stations
        """
        if not self.plot:
            return
        # Pick_Function.m:250
        self.stations = sorted(stations)
        N = len(stations)
        self.ax.set_title('Phase 1: Calculate global window')
        self.ax.set_xlabel(starttime.strftime('%Y-%m-%d'))
        self.ax.set_xlim(starttime.datetime, endtime.datetime)
        self.ax.set_ylim(-0.5, N + 0.5)
        # self.ax.set_ylim(-0.5, N-0.5)
        self.ax.set_yticks(np.arange(0, N))
        self.ax.set_yticklabels(stations)
        self.ax.set_xlabel(starttime.strftime('%Y-%m-%d'))
        plt.draw()
        self.fig.canvas.draw_idle()
        plt.show(block=False)
        plt.pause(0.001)

    def plot_trace(self, trace, station, candidates):
        """
        Plot one trace
        :param trace: trace to plot
        :param station: station name
        :param candidates: candidate picks
        """
        # Pick_Function.m: ~160?
        if not self.plot:
            return
        i_sta = self.stations.index(station)
        norm_trace = trace.copy().detrend().normalize()
        self.ax.plot_date(norm_trace.times(type="matplotlib"),
                          (0.5 * norm_trace.data) + i_sta,
                          fmt='-', label=trace.stats.station)
        self.ax.vlines([x.time.matplotlib_date for x in candidates],
                       i_sta + 0.45, i_sta - 0.45, color='k')
        # plt.draw()
        # self.fig.canvas.draw_idle()
        # plt.show(block=False)
        # plt.pause(0.001)

    def plot_pickbounds(self, rect_start_time, rect_end_time):
        """
        Plot the global pick bounds

        :param rect_start_time: global pick window start time
        :param rect_end_time: global pick window end time
        """
        if not self.plot:
            return
        N = len(self.stations)
        width = rect_end_time - rect_start_time
        patch = Rectangle((rect_start_time.matplotlib_date, -0.5),
                          width=width/86400, height=N + 1, color='lightblue',
                          alpha=0.3, edgecolor=None)
        self.ax.axvline(rect_start_time.matplotlib_date)
        self.ax.axvline(rect_end_time.matplotlib_date)
        self.ax.add_patch(patch)
        # plt.draw()
        self.fig.canvas.draw_idle()
        plt.show(block=False)
        plt.pause(0.001)


class Picks_Window():
    """
    Picks window Plotter

    Shows picks in the selected pick window, for all stations
    zorders: 0: windows: 1: candidates (wide, alpha=0.5),
             2: initial p-s candidates, (wide, alpha=0.5), 3: trace,
             4: cluster candidates, 5: final picks
    """
    def __init__(self, plot=True):
        self.plot = plot
        if not plot:
            self.fig, self.ax = None, None
        else:
            self.fig, self.ax = plt.subplots(num="PSPicker: all picks",
                                             clear=True)
        self.stations = []  # ordered list of stations

    def setup(self, starttime, endtime, stations):
        """
        :param starttime: plot start time
        :param endtime: plot end time
        :param stations: sorted list of stations
        """
        if not self.plot:
            return
        # Pick_Function.m:250
        self.stations = sorted(stations)
        N = len(stations)
        # fig.clf()
        self.ax.set_title('Phase 3: Select final picks using Association')
        # ax.set_xlim(starttime.matplotlib_date, endtime.matplotlib_date)
        self.ax.set_xlim(starttime.datetime, endtime.datetime)
        self.ax.set_ylim(-0.5, N-0.5)
        self.ax.set_yticks(np.arange(0, N))
        self.ax.set_yticklabels(stations)
        self.ax.set_xlabel(starttime.strftime('%Y-%m-%d'))
        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)

    def plot_traces_candidates(self, traces, c_P, c_S, candidates, station,
                               assoc=Associator):
        """
        Plot P trace(s) and candidates for one station

        :param traces: Stream of traces to plot
        :param c_P: P PickCandidate or None
        :param c_S: S PickCandidate  or None
        :param candidates: list of PickCandidates
        :param station: station name
        :param assoc: Associator object
        """
        if not self.plot:
            return
        ax = self.ax
        i_sta = self.stations.index(station)
        # Pick_Function.m:601
        for tr in traces:
            norm_trace = tr.copy().normalize()
            ax.plot_date(norm_trace.times(type="matplotlib"),
                         norm_trace.data + i_sta,
                         color='gray', ls='-', marker=None, zorder=3)
        for cand in candidates:
            ax.vlines(cand.time.matplotlib_date, i_sta-0.5, i_sta+0.5,
                      color='gray', lw=5, alpha=0.5, zorder=1)
        if c_P is not None:
            ax.vlines(c_P.time.matplotlib_date, i_sta-0.5, i_sta+0.5,
                      color='b', lw=5, alpha=0.5, zorder=2)
        if c_S is not None:
            ax.vlines(c_S.time.matplotlib_date, i_sta-0.5, i_sta+0.5,
                      color='r', lw=5, alpha=0.5, zorder=2)
        if c_S is not None and c_P is not None:
            o_time = assoc.calc_origin_time(c_P.time, c_S.time)
            ax.plot(o_time.matplotlib_date, i_sta, 'k+')
        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)

    def plot_picks(self, picks, t_begin, assoc):
        """
        Plot initial and final P and S picks for all stations

        :picks: all P and S PickCandidates
        :param t_begin: begin time for all plots
        :param assoc: Associator object
        """
        if not self.plot:
            return
        # Pick_Function.m:811
        p_picks = [p for p in picks if p.phase_guess == 'P']
        s_picks = [p for p in picks if p.phase_guess == 'S']
        self._add_picks(p_picks, 'b')
        self._add_picks(s_picks, 'r')
        self._add_cluster_rectangle(assoc.p_cluster, 'b', '#a0a0ff')
        self._add_cluster_rectangle(assoc.s_cluster, 'r', '#ffa0a0')
        self._add_cluster_rectangle(assoc.o_cluster, 'g', '#a0ffa0')
        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)

    def _add_picks(self, picks, color):
        """
        Plot picks for one station as solid lines

        :parm picks: list of Picks
        :param color: color for picks
        """
        if len(picks) == 0:
            return
        for pick in picks:
            j = self.stations.index(pick.station)
            self.ax.vlines(pick.time.datetime, j - 0.5, j + 0.5, colors=color,
                           zorder=5)
        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)

    def _add_cluster_rectangle(self, cluster, color, face_color):
        if len(cluster) == 0:
            return
        for sta, t in cluster.items():
            self.ax.plot(t.matplotlib_date, self.stations.index(sta),
                         color=color, marker='x', ls=None, zorder=4)
        t = [x.matplotlib_date for x in cluster.values()]
        # log(cluster, 'debug')
        # log(min(t), 'debug')
        # log(max(t), 'debug')
        p = Rectangle((min(t), -0.5), width=max(t) - min(t),
                      height=len(self.stations), fc=face_color, ec=color,
                      alpha=0.5, zorder=0)
        self.ax.add_patch(p)


class Station_Window():
    """
    Station window Plotter

    Detailed plot of picks, data, kurtosis, energy and SNR for one station
    Figures 3+: Station plots, each one divided in four:
        1) Picks
        2) Data
        3) Kurtosis
        4) Energy and signal-to-noise level
        5) Polarity
    """
    def __init__(self, plot=True):
        """
        Initialize station window object (but not the window)
        """
        self.plot = plot
        self.fig = None
        self.ax_picks = None
        self.ax_data = None
        self.ax_kurt = None
        self.ax_snr = None
        self.ax_pol = None

    def setup(self, trace):
        """
        Plot pick phases on Figure 3+

        :param trace: trace to plot
        """
        if not self.plot:
            return
        # Pick_Function.m:334
        name = trace.stats.station
        self.fig, axs = plt.subplots(5, 1, sharex=True, num=name, clear=True)
        self.fig.subplots_adjust(hspace=0)
        self.ax_picks = axs[0]
        self.ax_cand = axs[1]
        self.ax_kurt = axs[2]
        self.ax_snr = axs[3]
        self.ax_pol = axs[4]

        # fig.clf()
        self.ax_picks.set_title('Phase 2: Pick phases (station {})'
                                .format(trace.stats.station))

    def plot_data(self, starttime, endtime, snr_quality_thresholds,
                  trace, energy, kurto, dip_rect,
                  dip_thresh_P, dip_thresh_S):
        """
        Plot the waveforms, energy, kurtosis, etc

        :param starttime: starttime for all plots
        :param endtime: endtime for all plots
        :param snr_quality_thresholds: list of SNR quality thresholds
        :param trace: data trace to plot
        :param energy: WaveformEnergy object
        :param kurto: Kurtosis object
        # :param kurto_mean: mean Kurtosis
        # :kurto_grad: list of cumulative kurtosis gradient Traces, for
        #              different smoothings
        :dip_rect: dip-rectilinearity trace or None
        :dip_thresh_P: P-wave dip-rect threshold
        :dip_thresh_S: S-wave dip-rect threshold
        """
        if not self.plot:
            return
        # Pick_Function.m:443
        stats = kurto.mean_kurtosis.stats
        tmin, tmax = stats.starttime, stats.endtime
        # tmin, tmax = kurto_mean.stats.starttime, kurto_mean.stats.endtime
        self.kmin = np.min(kurto.mean_kurtosis.data)
        self.kmax = np.max(kurto.mean_kurtosis.data)

        # Plot the subplots
        self.ax_picks.set_xlim(starttime.datetime, endtime.datetime)
        self.fig.tight_layout()   # (to avoid clipping right labels)
        self._plot_picks(self.ax_picks, trace.slice(tmin, tmax))
        self._plot_cand(self.ax_cand, trace.slice(tmin, tmax))
        self._plot_kurt(self.ax_kurt, kurto)
        # self._plot_kurt(self.ax_kurt, kurto_mean, kurto_grad_stream)
        self._plot_snr(self.ax_snr, energy.slice(tmin, tmax),
                       snr_quality_thresholds)
        self._plot_pol(self.ax_pol, dip_rect, dip_thresh_P, dip_thresh_S)

        self.fig.tight_layout()   # (to avoid clipping right labels)
        self.fig.subplots_adjust(hspace=0.1)

        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)

    @staticmethod
    def _plot_picks(ax, trace):
        dmin, dmax = np.nanmin(trace.data), np.nanmax(trace.data)
        # patch = Rectangle((tmin.matplotlib_date, dmin),
        #                   width=(tmax - tmin)/86400, height=dmax - dmin,
        #                   color=[.9, .9, .9], edgecolor=None)
        ax.plot(trace.times(type="matplotlib"), trace.data, 'k')
        # ax.add_patch(patch)
        ax.set_ylim(dmin, dmax)
        ax.set_ylabel('Picks')

    @staticmethod
    def _plot_cand(ax, trace):
        pass

    @staticmethod
    def _plot_kurt(ax, kurto):
        kurto_mean = kurto.mean_kurtosis
        kurto_cum_mean = kurto.mean_cumulative_kurtosis
        smoov = kurto.params.extrema_smoothings
        tmin, tmax = kurto_mean.stats.starttime, kurto_mean.stats.endtime
        kurto_grad = Stream(kurto.kurto_gradients).slice(tmin, tmax)

        ax.plot(kurto_mean.times(type='matplotlib'), kurto_mean.data, 'b',
                label='mean')
        ax.plot(kurto_cum_mean.times(type='matplotlib'), kurto_cum_mean.data,
                'b--', label='cumulative')
        ax.set_ylabel('Kurtosis', color='b')
        ax.tick_params(axis='y', labelcolor='b')
        ax.legend(frameon=False, fontsize='xx-small')   # loc='center right'
        # ax.set_xticklabels([])
        axb = ax.twinx()  # Create a twin axis with a different y-scale
        axb.plot(kurto_grad[0].times(type='matplotlib'),
                 kurto_grad[0].data, 'r--',
                 label=f'{max(smoov):g}-s smoothing')
        axb.plot(kurto_grad[-1].times(type='matplotlib'),
                 kurto_grad[-1].data, 'r',
                 label=f'{min(smoov):g}-s smoothing')
        axb.set_ylabel('Extrema', color='r')
        axb.tick_params(axis='y', labelcolor='r')
        axb.legend(frameon=False, fontsize='xx-small')  # loc='center right'

    @staticmethod
    def _plot_snr(ax, energy, snr_quality_thresholds):
        nrgdB = energy.nrg.copy()
        nrgdB.data = 20 * np.log10(nrgdB.data)
        # ax.plot_date(energy.nrg.times(type='matplotlib'), energy.nrg.data,
        #              'r', zorder=0., alpha=0.5)
        ax.semilogy(energy.nrg.times(type='matplotlib'), energy.nrg.data,
                    'r', zorder=0., alpha=0.5)
        ax.set_ylabel('Energy', color='r')
        ax.tick_params(axis='y', labelcolor='r')
        axb = ax.twinx()  # Create a twin axis with a different y-scale
        # axb.semilogy(energy.snr.times(type='matplotlib'), energy.snr.data,
        #               'b', zorder=1.)
        # axb.set_ylim(bottom=1.)
        axb.plot(energy.snr.times(type='matplotlib'), energy.snr.data,
                 'b', zorder=1.)
        axb.set_ylim(bottom=0, top=1.5*max(snr_quality_thresholds))
        if energy.snr_threshold is not None:
            axb.axhline(energy.snr_threshold, color='b', ls='--', zorder=0.5)
        for thresh in snr_quality_thresholds:
            axb.axhline(thresh, color='b', ls='--', lw=0.1, zorder=0.5)
        axb.set_ylabel('SNR', color='b')
        axb.tick_params(axis='y', labelcolor='b')

    @staticmethod
    def _plot_pol(ax, trace, dip_thresh_P, dip_thresh_S):
        """
        :param dip_thresh_P: minimum P-wave dip-rect
        :param dip_thresh_S: maximum S-wave dip-rect
        """
        if trace is not None:
            tr = trace.copy()
            tr.data[tr.data == 0] = np.nan
            ax.plot_date(tr.times(type='matplotlib'), tr.data, 'k')
            ax.axhline(0, color='k', ls='-', lw=0.1, zorder=0.5)
            ax.axhline(dip_thresh_P, color='b', ls='--', zorder=0.5)
            ax.axhline(dip_thresh_S, color='r', ls='--', zorder=0.5)
            ax.set_ylim(-1, 1)
        else:
            ax.text(0.5, 0.5, 'Polarity not calculated',
                    ha='center', va='center', transform=ax.transAxes)
        ax.set_ylabel('Dip-Rect')

    def candidates(self, candidates, DR_thresh_P, DR_thresh_S):
        """
        Add pick candidates to all axes

        :param candidates: list of PickCandidate
        """
        if not self.plot:
            return
        # Extrema picks are grey
        color = [0.8, 0.8, 0.8]
        # Plot candidate times across  all axes
        for c in candidates:
            self.ax_picks.axvline(c.time.matplotlib_date, color=color)
            self.ax_kurt.axvline(c.time.matplotlib_date, color=color)
            self.ax_snr.axvline(c.time.matplotlib_date, color=color)
            self.ax_pol.axvline(c.time.matplotlib_date, color=color)

        # Detail candidates on dedicated axis
        ax = self.ax_cand
        axb = ax.twinx()  # Create a twin axis with a different y-scale
        ax.stem([x.time.matplotlib_date for x in candidates],
                [x.picker_value for x in candidates], markerfmt='ro',
                linefmt='r-', basefmt='r:', use_line_collection=True)
        ax.set_ylabel('picker_value', color='r')
        ax.tick_params(axis='y', labelcolor='r')
        log('Candidates:', 'debug')
        for x in candidates:
            log(f'  {x}', 'debug')
        axb.stem([x.time.matplotlib_date for x in candidates],
                 [x.DR for x in candidates], markerfmt='bx',
                 linefmt='b-', basefmt='b:', use_line_collection=True)
        axb.axhline(DR_thresh_P, color='b', ls='--')
        axb.axhline(DR_thresh_S, color='b', ls='--')
        axb.set_ylabel('dip-rect', color='b')
        axb.tick_params(axis='y', labelcolor='b')
        axb.set_ylim(-1, 1)
        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)

    def onsets(self, onset_P, onset_S, data_limits):
        """
        Plot onsets for one station, on pick and station windows

        :param onset_P: P PickCandidate or None
        :param onset_S: S PickCandidate or None
        :param data_limits: minimum and maximum data values
        """
        if not self.plot:
            return
        # Pick_Function.m:576
        # Station Window: add picks to subplots 1, 2 and 3
        if onset_P is not None:
            t = onset_P.time.matplotlib_date
            self.ax_picks.vlines(t, data_limits[0], data_limits[1], 'b')
            self.ax_kurt.vlines(t, self.kmin, self.kmax, 'b')
            # self.ax_extrem.vlines(t, data_limits[0], data_limits[1], 'b')
        if onset_S is not None:
            t = onset_S.time.matplotlib_date
            self.ax_picks.plot(t, data_limits[0], data_limits[1], 'r')
            # self.ax_extrem.plot(t, data_limits[0], data_limits[1], 'r')
            self.ax_kurt.plot(t, self.kmin, self.kmax, 'r')
        plt.draw()
        plt.show(block=False)
        plt.pause(0.001)



import copy

import numpy as np
from obspy.signal.util import smooth as obspy_smooth
from obspy.core.stream import Stream
from obspy.core import UTCDateTime

from .utils import smooth_filter
from .logger import log


class EnergySNR():
    """
    Energy and signal-to-noise ratio of a waveform Stream
    """
    def __init__(self, stream, params, plot=False):
        """
        calculate energy and signal-to-noise ratio of a seismic stream

        :param stream: group of related traces to calculate over
        :param params: SNRParameters object
        """
        assert isinstance(stream, Stream), "stream is not an obspy Stream"

        self.params = params
        self.snr_threshold = None
        self.nrg = self._calc_energy(stream)
        self.station = self._get_station(stream)
        self.snr = self._calc_snr()
        self.nrg.stats.channel = 'NRG'
        self.snr.stats.channel = 'SNR'

        if plot:
            snr_dB = self.snr.copy()
            snr_dB.data = 20*np.log10(snr_dB.data)
            snr_dB.stats.channel = 'SDB'
            Stream([stream[0], snr_dB, self.snr, self.nrg]).plot(
                equal_scale=False)

    def copy(self):
        return copy.copy(self)

    def deepcopy(self):
        return copy.deepcopy(self)

    @staticmethod
    def _calc_energy(stream):
        temp = stream.copy()
        for t in temp:
            t.data = np.power(t.data, 2)
        try:
            energy = temp.stack(npts_tol=1,
                                time_tol=1/temp[0].stats.sampling_rate)[0]
            if energy.stats.starttime.timestamp == 0:
                log('stacked traces are offset by more than one sample'
                    'setting starttime to stream[0].stats.starttime',
                    'error')
                energy.stats.starttime = stream[0].stats.starttime
        except ValueError as err:
            log(err, 'error')
            log('Slicing to same size', 'warning')
            log(temp, 'warning')
            last_start = UTCDateTime(max([t.stats.starttime.timestamp
                                          for t in temp]))
            first_end = UTCDateTime(min([t.stats.endtime.timestamp
                                         for t in temp]))
            temp.trim(last_start, first_end)
            log(temp, 'warning')
            energy = temp.stack(npts_tol=1)[0]
        energy.data = np.power(energy.data, 0.5)
        return energy

    @staticmethod
    def _get_station(stream):
        stations = list(set([t.stats.station for t in stream]))
        if len(stations) == 1:
            return stations[0]
        else:
            return "MULTIPLE STATIONS"

    def _calc_snr(self):
        """
        Return estimated signal to noise ratio

        :param energy: energy trace
        :returns: (M(t...t+window_after)/M(t-window_before,t))
        """
        sr = self.nrg.stats.sampling_rate
        half_noise_wind_samps = round(sr * self.params.noise_window / 2)
        half_signal_wind_samps = round(sr * self.params.signal_window / 2)
        self.params.noise_window = 2*half_noise_wind_samps/sr
        self.params.signal_window = 2*half_signal_wind_samps/sr
        w_noise = self.nrg.copy()
        w_signal = self.nrg.copy()
        # smooth using central moving average
        w_noise.data = obspy_smooth(self.nrg.data, half_noise_wind_samps)
        w_signal.data = obspy_smooth(self.nrg.data, half_signal_wind_samps)
        # Shift times so that noise is BEFORE reference time and signal AFTER
        w_noise.stats.starttime += self.params.noise_window/2
        w_signal.stats.starttime -= self.params.signal_window/2
        noise_start = w_noise.stats.starttime
        signal_end = w_signal.stats.endtime
        if noise_start >= signal_end:
            log("SNR noise_start >= signal_end, no SNR calculated"
                "(trace, signal, noise lengths = {.2g}, {.2g}, {.2g})".format(
                self.nrg.stats.npts / self.nrg.stats.sampling_rate,
                self.params.signal_window,
                self.params.noise_window), 'error')
            return None
        w_noise.trim(noise_start, signal_end)
        w_signal.trim(noise_start, signal_end)
        w_signal.data /= w_noise.data
        return w_signal

    def slice(self, starttime=None, endtime=None, nearest_sample=True):
        """
        Returns sliced view of the nrg and snr traces
        """
        new = self.copy()
        new.nrg = self.nrg.slice(starttime=starttime, endtime=endtime,
                                 nearest_sample=nearest_sample)
        new.snr = self.snr.slice(starttime=starttime, endtime=endtime,
                                 nearest_sample=nearest_sample)
        return new

    def is_trustworthy(self, n_smooth=100, debug=False):
        """
        Check if the signal can be trusted or not.

        Consider the signal trustworthy if the SNR crosses (from below to
        above) the specified threshold at least once and no more than
        SNR_threshold_crossings times
        :param n_smooth: length of moving average filter to apply before
            analysis
        :returns: is_trustworthy, info_text
        """
        # Pick_Function.m:380
        if self.snr.stats.npts == 0:
            log(f'station {self.station} self.snr has zero length', 'error')
            return False, 'error'
        snr_smooth = smooth_filter(self.snr, n_smooth)
        if snr_smooth is None:
            log('Could not smooth_filter self.snr for station "{}"'
                .format(self.station), 'error')
            log(self.snr, 'error')
            return False, 'error'
        self._set_snr_threshold(snr_smooth)
        sign_change = np.diff(np.sign(snr_smooth.data - self.snr_threshold))
        crossings = len(sign_change[sign_change == 2])
        if debug:
            snr_smooth.plot()
        max_cross = self.params.max_threshold_crossings
        trustworthy = (crossings > 0 and crossings <= max_cross)
        if not trustworthy:
            s = 'not trustworthy: '
            if crossings == 0:
                s += f'never crossed the threshold ({self.snr_threshold})'
            else:
                s += f'crossed the threshold ({self.snr_threshold}) '
                s += f'{crossings:d} times (> {max_cross:d})'
        else:
            s = "trustworthy"
        return (crossings > 0
                and crossings <= self.params.max_threshold_crossings), s

    def _set_snr_threshold(self, snr_smooth):
        """
        Calculate the signal-to-noise threshold value

        :param snr_smooth: trace of smoothed signal-to-noise ratio
        """
        # Pick_Function.m:382
        tp = self.params.threshold_parameter
        min_threshold = min(self.params.quality_thresholds)
        if (tp > 0 and tp <= 1):
            threshold = 1 + tp * (np.nanmax(snr_smooth.data) - 1)
        else:
            assert tp < 0, f'Illegal SNR threshold_parameter value: {tp:g}'
            threshold = -tp
        self.snr_threshold = max([threshold, min_threshold])


if __name__ == '__main__':
    pass



# Generated with SMOP  0.41-beta
# from smop.libsmop import *

# Standard libraries
from pathlib import Path
import shutil
import warnings
# import glob
# import warnings
# from logging import info
from datetime import datetime
# import sys

# Publicly available libraries
import numpy as np
from scipy import stats
from obspy.core import UTCDateTime
from obspy.core.event import Catalog as obspy_Catalog
# from obspy.core.inventory.response import PolesZerosResponseStage
from obspy.core.event.origin import Origin as obspy_Origin
from obspy.core.event import Event as obspy_Event
# from obspy.signal.invsim import simulate_seismometer
from obspy.io.nordic.core import read_nordic
from obspy.core import read as obspy_read

# module libraries
from .parameters import (PickerParameters, PickerRunParameters,
                         PickerStationParameters)
from .kurtosis import Kurtosis
from .energy_snr import EnergySNR
from .polarity import Polarity
from .pick_candidate import PickCandidate
from .associator import Associator
from .plotter import Plotter
from .local_amplitude import LocalAmplitude
from .utils import (select_traces, smooth_filter, picks_ps_times)
from .logger import setup_log, log
from .timer import Timer

warnings.filterwarnings("ignore",
                        message="Lines of type I have not been implemented "
                                "yet, please submit a development request")


class PSPicker():
    """
    Pick P and S arrivals on multiple stations using the Kurtosis and
    do a very basic cluster-based association
    """
    def __init__(self, parm_file, wav_base_path, database_path_in,
                 database_path_out='./Sfile_directory',
                 database_format='NORDIC'):
        """
        :param parm_file: path/name of the parameter file
        :param wav_base_path: absolute basepath to the waveform files
            (just before the YEAR/MONTH subdirectories)
        :param database_path_in: absolute basepath to the database/catalog
            file(s) (just before the YEAR/MONTH subdirectories)
        :param database_path_out: path to output database files
        :param database_format:
            'NORDIC': assume waveform files and database files are named using
                SEISAN conventions and located in YEAR/MONTH subdirectories
                under wav_base_path and database_path_in, respectively
        """
        self.parm_file = parm_file
        self.wav_base_path = Path(wav_base_path)
        self.database_path_in = Path(database_path_in)
        self.database_path_out = Path(database_path_out)
        if not self.database_path_out.is_dir():
            assert not self.database_path_out.exists()
            self.database_path_out.mkdir()
        # self.database_filename = None
        self.param = PickerParameters.from_yaml_file(parm_file)
        self.debug_plots = False
        self.run = None
        self.assoc = None
        # self.log_level = None

    def __str__(self):
        """
        """
        str = "PSPicker\n"
        str += f"    wav_base: {self.wav_base_path}\n"
        str += f"    output db directory: {self.database_path_out}\n"
        str += f"    parameters: {self.param}\n"
        return str

    def run_many(self, start_date, end_date, plot_global=False,
                 plot_stations=False, ignore_fails=True, log_level='info'):
        """
        Loops over events in a date range

        :param start_date: first data to process
        :param end_date: last data to process
        :type start_date, end_date: str of format "YYYYMMDD" or "YYYYMMDDHHMM"
        :param plot_global: show global and overall pick plots
        :param plot_stations: show individual station plots
        :param ignore_fails: keep going if one run fails
        :param log_level: console log level (choices = 'debug', 'verbose',
            'info', 'warning', 'error', 'critical'), default='info'
        """
        setup_log(log_level)
        # self.log_level = log_level
        start_dt = self._split_date(start_date)
        end_dt = self._split_date(end_date)
        debug_fname = 'ps_picker_debug_{}-{}_run{}.txt'.format(
                start_date, end_date, datetime.today().strftime('%Y%m%d'))
        log('Running from {} to {}'.format(start_dt.strftime("%Y%m%d-%H%M"),
                                           end_dt.strftime("%Y%m%d-%H%M")))

        # Print parameter information
        log(str(self.param), 'verbose')

        def _date_match(y, m, d, ref):
            return y == ref.year and m == ref.month and d == ref.day

        for year in range(start_dt.year, end_dt.year + 1):
            first_month = 1
            last_month = 12
            if year == start_dt.year:
                first_month = start_dt.month
            if year == end_dt.year:
                last_month = end_dt.month
            for month in range(first_month, last_month + 1):
                first_day = 1
                last_day = 31
                if year == start_dt.year and month == start_dt.month:
                    first_day = start_dt.day
                    # first_hour = start_dt.hour
                    # first_minute = start_dt.minute
                if year == end_dt.year and month == end_dt.month:
                    last_day = end_dt.day
                for day in range(first_day, last_day + 1):
                    kwargs = {}
                    if _date_match(year, month, day, start_dt):
                        kwargs['first_hour'] = start_dt.hour
                        kwargs['first_minute'] = start_dt.minute
                    if _date_match(year, month, day, end_dt):
                        kwargs['last_hour'] = end_dt.hour
                        kwargs['last_minute'] = end_dt.minute
                    self._run_one_day(year, month, day, plot_global,
                                      plot_stations, ignore_fails,
                                      debug_fname, **kwargs)

    def run_one(self, database_filename, plot_global=True, plot_stations=False,
                assoc=None, log_level='verbose', debug_plots=None):
        """
        Picks P and S arrivals on one waveform, using the Kurtosis

        Information in the database file will be appended with the picks.
        :param database_filename: database file to read
        :param plot_global: show global and overall pick plots
        :param plot_stations: show individual station plots
        :param assoc: Associator object (used for multiple runs with same
            Associator)
        :param log_level: console log level (choices = 'debug', 'verbose',
            'info', 'warning', 'error', 'critical').  If None, do not setup log
        :param debug_plots: plot some "debugging" plots
        """
        # if not self.log_level:
        #     setup_log(log_level)
        #     self.log_level = log_level
        if log_level is not None:
            setup_log(log_level)
        if self.assoc is None:
            self.assoc = Associator(self.param.assoc)
        if debug_plots is not None:
            self.debug_plots = debug_plots
        log(f'running {database_filename}', 'debug')
        timer = Timer(logger=None)
        timer.start()
        # Run basic Kurtosis and assoc to find most likely window for picks
        # with Timer(text="Setup Plotter: {:0.4f}s"):
        plotter = Plotter(plot_global, plot_stations)

        # Read in data and select global pick window
        # with Timer(text="Read waveforms: {:0.4f}s"):
        st, wavefile = self._read_waveforms(
            self._full_nordic_database_filename(database_filename))
        # print(st.__str__(extended=True))
        if len(st) == 0:
            log('No data found in {wavefile}, referred by {database_filename}',
                'error')
            return
        sta_list = sorted(list(set([tr.stats.station for tr in st])))
        log('Read waveforms from stations {}'.format(', '.join(sta_list)),
            'verbose')
        # with Timer(text="Choose global window: {:0.4f}s"):
        cmaps, ft, lt = self._choose_global_window(st, plotter)
        _check_timelimits(st, ft, lt)
        self.run = PickerRunParameters(
            database_filename=database_filename, wavefile=wavefile,
            stream=st, channel_maps=cmaps, first_time=ft, last_time=lt)
        plotter.gw.plot_pickbounds(ft, lt)
        plotter.pw.setup(ft, lt, self.run.stations)

        # Pick on individual traces
        candidates, picks = [], []
        for sta, chan_map in self.run.channel_maps.items():
            # Reject stations not listed in parameter file
            if sta not in self.param.stations:
                log(f'{sta} not in self.param.stations, ignored', 'warning')
                continue
            # with Timer(text="Pick one station: {:0.4f}s"):
            p, c = self._pick_one_station(sta, chan_map, plotter)
            picks.extend(p)
            candidates.extend(c)

        # with Timer(text="Associate: {:0.4f}s"):
        picks = self.assoc.run(picks, candidates)
        plotter.pw.plot_picks(picks, self.loop.t_begin, self.assoc)
        # with Timer(text="Save picks: {:0.4f}s"):
        # log(f'picks = {picks}', 'debug')
        picks = PickCandidate.remove_duplicates(picks)
        obspy_pa = [x.to_obspy(self.run.channel_maps,
                               self.param.SNR.quality_thresholds)
                    for x in picks]
        obspy_picks = [x[0] for x in obspy_pa]
        obspy_arrivals = [x[1] for x in obspy_pa if x[1] is not None]
        # amplitudes, obspy_picks = self._calc_amplitudes(obspy_picks)
        amplitudes, amp_picks = self._calc_amplitudes(obspy_picks)
        obspy_picks.extend(amp_picks)
        self._save_event(obspy_picks, amplitudes, obspy_arrivals)
        elapsed_time = timer.stop()
        log('    {}: {:2d} Picks and {:2d} Amplitudes on {:2d} stations in '
            '{:0.2f} seconds'.format(database_filename,
                                     len(obspy_picks) - len(amplitudes),
                                     len(amplitudes), len(cmaps),
                                     elapsed_time))

    def _run_one_day(self, year, month, day, plot_global, plot_stations,
                     ignore_fails, debug_fname, first_hour=None,
                     first_minute=None, last_hour=None, last_minute=None):
        """Select and run events for one day"""
        log(f'Running {year}-{month}-{day}, {first_hour=}, '
            f'{first_minute=}, {last_hour=}, {last_minute=}', 'debug')
        db_path_in = self.database_path_in / f'{year:04d}' / f'{month:02d}'
        s_files = list(db_path_in.glob(f'{day:02d}-*.S*'))
        if len(s_files) > 0:
            if first_hour is not None or first_minute is not None:
                s_files = [f for f in s_files if self._nordic_fname_after(
                           f.name, first_hour, first_minute)]
            if last_hour is not None or last_minute is not None:
                s_files = [f for f in s_files if self._nordic_fname_before(
                           f.name, last_hour, last_minute)]
        s_files.sort()
        if len(s_files) > 0:
            log('Running {:d} events on {:04d}-{:02d}-{:02d}'.format(
                len(s_files), year, month, day))
            for s_file in s_files:
                log("   Running {}...".format(s_file), 'verbose')
                t = Timer(logger=None)
                t.start()
                try:
                    self.run_one(s_file, plot_global=plot_global,
                                 plot_stations=plot_stations, log_level=None)
                except Exception as err:
                    log(f'run_one() failed for {s_file}', 'critical')
                    log(err, 'error')
                    if not ignore_fails:
                        raise Exception(err)
                    log('copying original s-file to dest', 'info')
                    inf = s_file
                    outf = self.database_path_out / Path(s_file).name
                    log(f'{inf} to {outf}')
                    shutil.copyfile(inf, outf)
                # elapsed_time = t.stop()

    @staticmethod
    def _nordic_fname_after(f, hour, minute):
        """
        Returns True if the event is on or after the given hour-minute
        :param f: filename (str)
        :param first_hour:
        :param first_minute:
        """
        ev_hour, ev_minute = int(f[3:5]), int(f[5:7])
        if ev_hour > hour:
            return True
        elif ev_hour == hour and ev_minute >= minute:
            return True
        return False

    @staticmethod
    def _nordic_fname_before(f, hour, minute):
        """
        Returns True if the event is on or before the given hour-minute
        :param f: filename (str)
        :param hour:
        :param minutes:
        """
        ev_hour, ev_minute = int(f[3:5]), int(f[5:7])
        if ev_hour < hour:
            return True
        elif ev_hour == hour and ev_minute <= minute:
            return True
        return False

    @staticmethod
    def _split_date(the_date):
        """Splits a date string ("YYYYMMDD" or "YYMMDDHHMM)
        :returns: datetime object year, month, day, hour, minute
        """
        d = the_date
        assert isinstance(d, str), 'date is not a string'
        assert d.isnumeric(), 'date is not numeric'
        assert len(d) == 8 or len(d) == 12,\
            f'"date" ({d}) is {len(d)} numerals, should be 8 or 12'
        if len(d) == 8:
            return datetime(int(d[:4]), int(d[4:6]), int(d[6:]))
        if len(d) == 12:
            return datetime(int(d[:4]), int(d[4:6]), int(d[6:8]), int(d[8:10]),
                            int(d[10:]))

    def _pick_one_station(self, station_name, chan_map, plotter):
        """
        Calculate picks and candidates for one station

        :param station_name: station name
        :param chan_map: ChannelMap object for the station
        :param plotter: Plotter object
        """
        # make shortened reference to often-used station_parameters
        # with Timer(text="  pick_one_station(): setup {:0.4f}s"):
        station_params = self.param.station_parameters[station_name]
        self.loop = PickerStationParameters(station=station_name,
                                            station_params=station_params,
                                            channel_map=chan_map,
                                            stream=self.run.stream)
        plotter.sw.setup(self.loop.datP[0])
        c_P, c_S, candidates = None, None, []

        # SNR analysis
        # with Timer(text="  pick_one_station(): SNR {:0.4f}s"):
        datS_filt = self.loop.datS.copy().filter(
            'bandpass', corners=3,
            freqmin=station_params.energy_frequency_band[0],
            freqmax=station_params.energy_frequency_band[1])
        energy = EnergySNR(datS_filt, self.param.SNR, plot=self.debug_plots)
        trust, message = energy.slice(self.run.first_time,
                                      self.run.last_time).is_trustworthy()
        log(f"{station_name}: SNR {message}", 'verbose')
        if trust:
            # with Timer(text="  pick_one_station(): Kurtosis {:0.4f}s"):
            c_P, c_S, kurt, candidates = self._run_Kurtosis(energy, plotter)
            for c in candidates:
                c.station = station_name

            # Verify phases using Polarity analysis
            DR = None
            # with Timer(text="  pick_one_station(): Polarity analysis
            #      {:0.4f}s"):
            if station_params.use_polarity and (len(datS_filt) == 3):
                c_P, c_S, DR, candidates = self._polarity_analysis(
                        c_P, c_S, candidates, datS_filt)
            plotter.sw.plot_data(self.run.first_time, self.run.last_time,
                                 self.param.SNR.quality_thresholds,
                                 self.loop.datP[0],
                                 energy,
                                 kurt,
                                 DR,
                                 self.param.polarity.DR_threshold_P,
                                 self.param.polarity.DR_threshold_S)
            if len(candidates) > 0:
                plotter.sw.candidates(candidates,
                                      self.param.polarity.DR_threshold_P,
                                      self.param.polarity.DR_threshold_S)

        plotter.pw.plot_traces_candidates(self.loop.datP, c_P, c_S,
                                          candidates, station_name, self.assoc)
        plotter.sw.onsets(c_P, c_S, self.loop.data_limits)

        new_picks = self._make_picks(c_P, c_S)
        return new_picks, candidates

    def _read_waveforms(self, database_filename, format='NORDIC'):
        if format == 'NORDIC':
            full_wavefile = self._get_nordic_wavefile_name(database_filename)
        else:
            raise NameError(f'type {type} not implemented')
        stream = obspy_read(full_wavefile, 'MSEED')
        # get rid of bad last sample in some streams, and detrend
        for tr in stream:
            tr.data = tr.data[:-10]
            tr.detrend(type='demean')
        return stream, full_wavefile

    def _get_nordic_wavefile_name(self, database_filename):
        log(f'database filename = {database_filename}', 'verbose')
        cat, wav_names = read_nordic(database_filename, return_wavnames=True)
        assert len(wav_names) == 1, 'More than one wav_name in database file'
        pts = wav_names[0][0].split('-')
        full_wav_name = self.wav_base_path / pts[0] / pts[1] / wav_names[0][0]
        return str(full_wav_name)

    def _full_nordic_database_filename(self, filename):
        """
        Look for a NORDIC database file and return its path

        Looks in local directory, then self.database_path_in, then
        self.database_path_in/YEAR/month

        :param filename: name of the file to look FormatControl
        :type filename: str or pathlib.Path
        """
        # In local directory
        if Path(filename).is_file():
            return filename
        # In database directory
        fullname = self.database_path_in / filename
        if fullname.is_file():
            return fullname
        # In database/year/month directory
        a = filename.split('.')[-1]
        year = a[1:5]
        month = a[5:]
        fullname = self.database_path_in / year / month / filename
        if fullname.is_file():
            return fullname
        else:
            raise NameError(f'database file "{filename}" not found')

    def _choose_global_window(self, stream, plotter):
        """
        Choose the global pick window

        :param stream: all the data read in
        :param plotter: Plotter object
        """
        t_begin = min([t.stats.starttime for t in stream])
        t_end = max([t.stats.endtime for t in stream])
        chan_maps = select_traces(stream, self.param.channel_mapping_rules)
        plotter.gw.setup(t_begin, t_end, [s for s in chan_maps.keys()])
        log(self._channel_maps_str(chan_maps), 'verbose')
        distri, chan_maps = self._gw_get_distri(stream, chan_maps, plotter)
        ft, lt, distri = self._gw_set_window(t_begin, t_end, distri)
        log(f'Global window bounds: {ft} to {lt}', 'verbose')
        return chan_maps, ft, lt

    def _channel_maps_str(self, channel_maps):
        s = 'Channel Map:\n'
        first_iter = True
        for key, v in channel_maps.items():
            if first_iter:
                s += f'{"Station":8s}|{v.__str__(format="table_header")}\n'
                first_iter = False
            s += f'{key:8s}|{v.__str__(format="table_row")}\n'
        return s

    def _gw_get_distri(self, stream, channel_maps, plotter, n_smooth=15):
        """
        Get overall pick distribution (and remove flat-lined stations)

        Finding the Global minimum on all the stations
        Filtering parameters + first window + smooth

        :param stream: all traces
        :param channel_maps: mapping of channel names to components
        :param plotter: the plotter object
        :n_smooth: samples to smooth kurtosis over
        :returns: overall_distribution of extrema, channel_maps, plotter
        """
        # Pick_Function.m:134
        p = self.param
        overall_distri = []
        rm_stations = []
        for station, channel_map in channel_maps.items():
            trace = stream.select(id=channel_map.Z)[0]
            assert station == trace.stats.station,\
                'trace station ({}) != iteration station ({})'.format(
                    trace.stats.station, station)
            if np.all(np.diff(trace.data) == 0):
                log(f'Station {station} flat-lined, ignoring', 'warning')
                rm_stations.append(station)
                continue
            p.gw.kurtosis.n_smooth = n_smooth
            k = Kurtosis(p.gw.kurtosis)
            candidates = k.pick_trace(trace, p.gw.max_candidates)
            for x in candidates:
                x.station = station
            overall_distri.extend([x.time for x in candidates])
            plotter.gw.plot_trace(trace, station, candidates)

        # REMOVE PROBLEM STATIONS (if necessary)
        channel_maps = {s: v for s, v in channel_maps.items()
                        if s not in rm_stations}
        return overall_distri, channel_maps

    def _gw_set_window(self, t_begin, t_end, overall_distri):
        """
        Define size of new analysis window

        :param t_begin: reference starttime for all traces
        :param t_end: end of all traces
        :param overall_distri: array of extrema on all stations (UTCDateTimes)
        :returns: first_time, last_time, overall_distri
        """
        # Pick_Function.m:204
        # max_offset is data length * gw_end_cutoff
        max_time = t_begin + self.param.gw.end_cutoff * (t_end - t_begin)
        # Cut down picks to those within global bounds
        overall_distri = [t for t in overall_distri if t <= max_time]
        min_global = UTCDateTime(center_distri(
            [t.timestamp for t in overall_distri], self.param.gw.distri_secs))

        first_time = max(min_global + self.param.gw.offsets[0], t_begin)
        last_time = min(min_global + self.param.gw.offsets[1], t_end)
        # if first_time < t_begin:
        #     first_time = t_begin
        # if last_time >= t_end:
        #     last_time = t_end
        return first_time, last_time, overall_distri

    def _run_Kurtosis(self, energy, plotter, debug=False):
        """
        calculate extrema and estimate P and S onsets using the Kurtosis

        :param energy: EnergySNR object
        :param plotter: Plotter object
        :returns:
            c_P: P PickCandidate
            c_S: S PickCandidate
            kurtosis: Kurtosis object
            candidates: list of all PickCandidtes
        """
        first_time, last_time = self._refine_pick_window(energy.nrg)
        if len(self.loop.datP) > 1:
            warnings.warn('Only working on first trace in datP')
        k = Kurtosis(self.loop.station_params.kurtosis)
        candidates = k.pick_trace(self.loop.datP[0],
                                  self.loop.station_params.max_candidates,
                                  first_time, last_time)
        #  Trace.times('utcdatetime') takes 0.3s per call!
        times = energy.snr.times('timestamp')
        imax = energy.snr.stats.npts - 1
        for c in candidates:
            c.snr = energy.snr.data[min(times.searchsorted(c.time.timestamp),
                                        imax)]

        # calculate picks using only candidates with SNR above min threshold
        strong_candidates = [x for x in candidates
                             if x.snr > min(self.param.SNR.quality_thresholds)]
        c_P, c_S = self._calc_follows(strong_candidates)
        return c_P, c_S, k, candidates

    def _refine_pick_window(self, energy):
        """
        Refine pick window to account for analysis windows?

        Seems like a convoluted way to reduce the time window by
        the longest kurtosis window length + the station energy window?
        I mean, why are we even smoothing the energy window?
        :param energy: energy trace
        """
        if self.loop.station_params.energy_window == 0:
            return (self.run.first_time,
                    self.run.last_time)

        energy_smooth = smooth_filter(energy, 50)
        energy_smooth.trim(self.run.first_time, self.run.last_time)
        starttime = energy_smooth.stats.starttime
        sr = energy_smooth.stats.sampling_rate
        # print(energy_smooth, type(energy_smooth), energy_smooth.data)
        ind_max = np.nanargmax(energy_smooth.data)
        last_sample = ind_max.copy()
        max_kurto_wind = np.max(self.loop.station_params.kurtosis
                                                        .window_lengths)
        max_precursor = np.floor(
            sr * (self.loop.station_params.energy_window + max_kurto_wind))
        first_sample = ind_max - max_precursor
        if first_sample < 0:
            first_sample = 0
            last_sample = max_precursor
        # Pick_Function.m:417
        return starttime + first_sample/sr, starttime + last_sample/sr

    def _polarity_analysis(self, c_P, c_S, candidates, datS_filtered):
        """
        Return P and S onsets from candidates based on signal polarity

        :param c_P: P-wave PickCandidate
        :param c_S: S-wave PickCandidate
        :param candidates: list of PickCandidate
        :param datS_filtered: stream of 3 traces used for S picking, filtered
            and with Z last
        :returns:
            c_P: PickCandidate P
            c_S: PickCandidate S
            DR: Dip-rectilinearity trace
            candidates: candidates with DR values added
        """
        if len(candidates) == 0:
            return None, None, None, candidates
        pol = Polarity(datS_filtered, params=self.param.polarity,
                       zcomponents=self.param.channel_mapping_rules.compZ,
                       ncomponents=self.param.channel_mapping_rules.compN,
                       ecomponents=self.param.channel_mapping_rules.compE)
        DR = pol.calc_dip_rect([c.time for c in candidates])
        if DR is None:
            log("DR not returned, keeping input picks", "debug")
            return c_P, c_S, DR, candidates

        for c in candidates:
            window = DR.slice(
                c.time, c.time + self.param.polarity.calculate_window).data
            c.DR = np.max(np.abs(window)) * np.sign(np.mean(window))
        c_P, c_S = None, None
        if len(candidates) == 1:
            c = candidates[0]
            if c.DR >= self.param.polarity.DR_threshold_P:
                c_P = c
            elif c.DR <= self.param.polarity.DR_threshold_S:
                c_S = c
        elif len(candidates) == 2:
            if candidates[0].time < candidates[1].time:
                cFirst, cSecond = candidates[0], candidates[1]
            else:
                cFirst, cSecond = candidates[1], candidates[0]
            if cFirst.DR >= self.param.polarity.DR_threshold_P:
                c_P = cFirst
            if cSecond.DR <= self.param.polarity.DR_threshold_S:
                c_S = cSecond
            if c_P is None or c_S is None:
                if cFirst.DR > cSecond.DR:
                    c_P = cFirst
                    c_S = cSecond
        log(f'polarity-verified c_P={c_P}, c_S={c_S}', 'verbose')
        return c_P, c_S, DR, candidates

    def _calc_follows(self, candidates):
        """
        returns offsets depending on number of extrema to follow

        :param candidates: list of PickCandidate
        :returns candidate_P, candidate_S
        :rtype: PickCandidate, PickCandidate
        """
        # Pick_Function.m:507
        # eliminate extrema whose snr is less than SNR.thresh
        max_candidates = self.loop.station_params.max_candidates
        # assert max_candidates in (1, 2), 'max_candidates is not 1 or 2'
        if len(candidates) == 0:
            return None, None
        if max_candidates == 1:
            return candidates[0], None
        else:
            # log(extrema, level='debug')
            candidates.sort(key=lambda x: x.time)
            if len(candidates) == 1:
                return candidates[0], None
            else:
                return candidates[0], candidates[1]

    def _make_picks(self, candidate_P, candidate_S):
        """
        Put onset_P and onset_S into list of Picks

        :param onset_P: PickCandidate for P
        :param onset_S: PickCandidate for S
        :param snr: signal-to-noise levels Trace
        """
        picks = []
        if candidate_P is not None:
            candidate_P.phase_guess = 'P'
            picks.append(candidate_P)
        if candidate_S is not None:
            candidate_S.phase_guess = 'S'
            picks.append(candidate_S)
        return picks

    def _calc_amplitudes(self, picks):
        amplitudes = []
        amp_picks = []
        stations = list(set([p.waveform_id.station_code for p in picks]))
        for station in stations:
            sta_picks = [p.copy() for p in picks
                         if p.waveform_id.station_code == station]
            # Inefficient way to get dat_noH
            temp = PickerStationParameters(
                station=station,
                station_params=self.param.station_parameters[station],
                channel_map=self.run.channel_maps[station],
                stream=self.run.stream)
            la = LocalAmplitude(
                temp.dat_noH, sta_picks,
                self.param.station_parameters[station].resp_file,
                self.param.response_file_type)
            # log(la, 'debug')
            amp, pick = la.get_iaml(method='wood_calc')
            if amp is not None:
                amplitudes.append(amp)
                amp_picks.append(pick)
        return amplitudes, amp_picks

    @staticmethod
    def save_nordic_event(picks, origin_time, filepath, filename,
                          amplitudes=[], arrivals=[], wavefiles=None,
                          evtype='L', debug=False):
        """
        Save event to NORDIC file

        Arrivals are used to fix the pick weighting: curently direct mapping,
        in version 2.0 will use inverse mapping which correponds to the
        difference between QuakeML and Nordic pick weights
        :param picks: event picks (including amplitude picks)
        :type picks: list
        :param origin_time: event origin time
        :type origin_time: UTCDateTime
        :param amplitudes: event amplitudes
        :type amplitudes: list
        :param arrivals: event arrivals
        :type arrivals: list
        :param wavefiles: list of waveform files corresponding to this event
        :param filepath: path to write to
        :param filename: name of file to write
        :param evtype: event type ('L', 'R' or 'D')
        """
        assert isinstance(picks, list)
        assert isinstance(amplitudes, list)
        assert isinstance(arrivals, list)
        # if len(picks) == 0:
        #     log('No picks to save!', 'warning')
        origin = obspy_Origin(time=origin_time, arrivals=arrivals)
        event = obspy_Event(
            event_type='earthquake',
            picks=picks,
            origins=[origin],
            amplitudes=amplitudes)
        log(event, 'verbose')
        cat = obspy_Catalog(events=[event])
        # How to change uncertainties to "0", "1", "2", "3"?
        # By creating an associated arrival and setting it's time_weight
        # to the appropriate number (which makes no sense because
        # a weight of zero should have no importance!)
        output_dbfile = Path(filepath) / filename
        cat.write(output_dbfile, format='NORDIC', evtype=evtype,
                  wavefiles=wavefiles, high_accuracy=True)

    def _save_event(self, picks, amplitudes=[], arrivals=[]):
        """
        Save event to NORDIC file
        """
        # Replaces a large section from Pick_Function.m 840-887
        if len(picks) == 0:
            o_time = self.run.first_time
        else:
            o_time = estimate_origin_time(picks)
        self.save_nordic_event(picks, o_time, self.database_path_out,
                               Path(self.run.database_filename).name,
                               amplitudes=amplitudes,
                               arrivals=arrivals,
                               wavefiles=[self.run.wavefile])

    def _write_debug_file(self, debug_fname, err, s_file):
        """
        Write debugging information after a failed run_one

        :param debug_fname: debug file name
        :param err: returned Exception error
        :param s_file: name of the s_file corresponding to this event
        """
        with open(debug_fname, 'a') as fid:
            fid.write('-'*60 + '\n')
            fid.write(f'{err}\n')
            fid.write('To reproduce the error, type:\n')
            fid.write(f'    picker.run_one("{s_file}")\n\n')


def _check_timelimits(st, ft, lt):
    """ Check if there are traces outside of the time limits """
    bad_traces = [tr for tr in st
                  if tr.stats.starttime > lt or tr.stats.endtime < ft]
    if len(bad_traces) > 0:
        stream = st.copy()
        stream.traces = bad_traces
        log('some traces are outside of the pick window:', 'error')
        log(stream, 'error')


def center_distri(v, win_size, n_steps=1000):
    """
    Return the center of the window containing the most values in an array

    :param v: array of values
    :param win_size: window size
    :param n_steps: number of values between min(v) and max(v) to test
    :returns: center of the distribution
    """
    if len(v) == 0:
        return []

    t = []
    s = []
    v_min, v_max = np.min(v), np.max(v)
    for tv in np.linspace(v_min, v_max, n_steps):
        a = tv - win_size / 2
        b = tv + win_size / 2
        in_range = np.logical_and(v > a, v < b)
        n_in_range = len(np.nonzero(in_range)[0])
        t.append(n_in_range)
        s.append(tv)
    return s[np.argmax(t)]


def estimate_origin_time(picks, vp_over_vs=1.7):
    """
    estimate EQ origin time based on pick times

    Uses P-S delays if possible. If not, return the earliest pick
    :param picks: list of obspy Pick objects
    :param vp_over_vs: assumed velocity ratio
    """
    origin_time = _average_ps_o_time(picks, vp_over_vs)
    if origin_time is not None:
        return origin_time
    else:
        times = list(sorted([p.time for p in picks]))
        # print(f'times={times}, picks={picks}', flush=True)
        return times[0]


def _average_ps_o_time(picks, vp_vs):
    """
    Find origin times for each P-S pair

    :param picks: list of obspy Pick objects
    :param vp_vs: assumed velocity ratio
    """
    o_times = []
    ps_delays, p_times = picks_ps_times(picks)
    if ps_delays is None:
        return None
    for ps, p in zip(ps_delays, p_times):
        o_times.append(Associator.calc_origin_time(p, p+ps, vpvs=vp_vs))
    # Throw out nans
    o_ts = [o.timestamp for o in o_times]
    # log(o_ts, 'debug')
    if np.any(np.isnan(o_ts)):
        log('there are o_times with timestamp = NaN!', 'error')
        o_ts = [o for o in o_ts if not np.isnan(o)]
    if len(o_ts) == 1:
        mean_timestamp = o_ts[0]
    elif len(o_ts) == 0:
        log('No valid origin times', 'warning')
        return None
    else:
        # Throw out values more than 3 std away
        zs = np.abs(stats.zscore([x for x in o_ts]))
        # log(zs, 'debug')
        if np.any(np.isnan(zs)):
            log("can't use z-values because they have NaNs "
                f"({len(o_ts)} origin_time(s))", 'warning')
            mean_timestamp = np.mean(o_ts)
        else:
            mean_timestamp = np.mean([o for o, z in zip(o_ts, zs) if z < 3])
    # log(f'mean_timestamp = {mean_timestamp}', 'debug')
    return UTCDateTime(mean_timestamp)


if __name__ == '__main__':
    pass



"""
Parameters used by ps_picker
"""
from .picker_parameters import PickerParameters
from .picker_run_parameters import PickerRunParameters
from .picker_station_parameters import PickerStationParameters
from .station_parameters import StationParameters
from .channel_mapping_rules import ChannelMappingRules

__all__ = ['PickerParameters', 'PickerRunParameters',
           'PickerStationParameters', 'StationParameters',
           'ChannelMappingRules']



"""
Pole-Zero representation of an instrument

Follows SEED Pole-Zero protocol for Analog Stages, radian format
(Seed Manual v2.4, Appendix C)

G(f) = S_d * A_0 * prod(s - z_n) / prod(s - p_n) = S_d * A_0 * Hp(s)
where s=j*2*pi*f

The corresponding attributes are:
    S_d: gain factor
    A_0: norm_factor
    G(f): gain at frequency f
    ref_gain = G(f0) = gain at the ref_freq. If there is a passband (a place
                       where the response is flat, ref_freq should be chosen
                       in this band)

The user can specify gain or ref_gain, not both
If norm_factor is not specified, it is automatically calculated so that
    A_0 * abs(Hp(s)) = 1 at f0
as per the SEED recommendation

The user can also change the input_units after creating the class, in which
case the poles, zeros and gain will  be modified if the starting and ending
units are in self.known_units().
"""
import numpy as np
import json
import copy
import warnings

from obspy.core.inventory import read_inventory
from obspy.core.inventory.response import PolesZerosResponseStage
from matplotlib import pyplot as plt


class PAZ():
    """
    Object-based implementation of Pole-Zero representation

    For representing a full instrument for simulations, etc
    """
    converter = {"m":       [1.,    2],
                 "m/s":     [1.,    1],
                 "m/s^2":   [1.,    0],
                 "m/s**2":  [1.,    0],
                 "mm":      [1.e-3, 2],
                 "mm/s":    [1.e-3, 1],
                 "mm/s^2":  [1.e-3, 0],
                 "mm/s**2": [1.e-3, 0],
                 "nm":      [1.e-9, 2],
                 "nm/s":    [1.e-9, 1],
                 "nm/s^2":  [1.e-9, 0],
                 "nm/s**2": [1.e-9, 0],
                 "gal":     [1.e-2, 0],   # 1 cm/s^2
                 "mgal":    [1.e-5, 0],
                 "ugal":    [1.e-8, 0]}

    def __init__(self, poles=[], zeros=[], ref_gain=None,
                 input_units='m/s', output_units='counts', ref_freq=1.,
                 gain=None, norm_factor=None, sampling_rate=None):
        """
        Can either specify gain or ref_gain, not both

        Frequency units are radians

        :param poles: poles (rad/s)
        :param zeros: zeros (rad/s)
        :param ref_gain: gain at ref_freq (should be w/in instrument passband)
        :param ref_freq: reference frequency (Hz) for passband gain
        :param input_units: units (usually physical) at the sensor entry
        :param output_units: units as stored
        :param gain: pole-zero gain (NOT passband gain)
        :param norm_fact: A0 normalization factor
        :param sampling_rate: channel sampling rate (used for response plot))
        """
        assert gain is None or ref_gain is None,\
            "both gain and ref_gain cannot be declared"
        if gain is None and ref_gain is None:
            ref_gain = 1.
        self.poles = np.array(poles)
        self.zeros = np.array(zeros)
        self._input_units = input_units
        self.output_units = output_units
        self.ref_freq = ref_freq
        if norm_factor is None:
            self._norm_factor = self.calc_norm_factor()
        else:
            self._norm_factor = norm_factor
        self.sampling_rate = sampling_rate
        if gain is not None:
            self.gain = gain
        else:
            self.gain = ref_gain * self.calc_norm_factor() / self.norm_factor

    @property
    def ref_gain(self):
        # return self.gain / self.norm_factor
        return self.gain * self.norm_factor / self.calc_norm_factor()

    @property
    def norm_factor(self):
        """ A0 normalization factor """
        return self._norm_factor

    @norm_factor.setter
    def norm_factor(self, value):
        """gain must change if norm factor changed"""
        self.gain *= self._norm_factor / value
        self._norm_factor = value

    @property
    def input_units(self):
        return self._input_units

    @input_units.setter
    def input_units(self, value):
        """
        Convert input (physical) units to a different value

        If newunits and self.input_units are one of the
        DEFINED UNITS (see _ref_meter()), changes gain, poles and zeros.
        """
        assert isinstance(value, str), 'value is not a character string'
        self._unit_conversion(value)
        self._input_units = value

    def _hp(self, freqs):
        """
        Return Hp(f) = prod(s - z_n) / prod(s - p_n)

        where s=j*2*pi*f
        """
        s = 2 * np.pi * np.array(freqs) * 1j
        hp = np.ones(s.shape, np.complex128)
        for p in self.poles:
            hp /= (s - p)
        for z in self.zeros:
            hp *= (s - z)
        return hp

    def __eq__(self, obj, places=2, verbose=False):
        """
        Return true if objects are equal to within places decimal places

        :param places: decimal place precision used for gain
        """
        err_msgs = []
        diff = abs((self.gain*self.norm_factor)
                   / (obj.gain*obj.norm_factor) - 1) * 10**places
        if diff > 1:
            err_msgs.append('different gain*A0 ()')
        if not len(self.zeros) == len(obj.zeros):
            err_msgs.append('n_zeros different')
        if not len(self.poles) == len(obj.poles):
            err_msgs.append('n_poles different')
        for z1, z2 in zip(self.zeros, obj.zeros):
            if not z1 == z2:
                err_msgs.append('zero different')
        for p1, p2 in zip(self.poles, obj.poles):
            if not p1 == p2:
                err_msgs.append('pole different')
        if not self.input_units.upper() == obj.input_units.upper():
            err_msgs.append('different input units')
        if not self.output_units.upper() == obj.output_units.upper():
            err_msgs.append('different output units')
        if not self.ref_freq == obj.ref_freq:
            err_msgs.append('different ref_freqs')
        if len(err_msgs) > 0:
            if verbose:
                print('PAZs NOT EQUAL:' + ','.join(err_msgs))
            return False
        return True

    def __str__(self):
        s = 'PAZ: '
        s += f'gain={self.gain:.4g}, '
        s += f'A0={self.norm_factor:.4g}, '
        s += f'poles={self.poles}, '
        s += f'zeros={self.zeros}, '
        s += f'(ref_gain={self.ref_gain:.3g} '
        s += f'{self.output_units}/{self.input_units} '
        s += f'at {self.ref_freq:.3g}Hz)'
        return s

    def copy(self):
        return copy.deepcopy(self)

    def get_response(self, freqs):
        """ return response for a given range of frequencies"""
        resp = self.gain * self.norm_factor * self._hp(freqs)
        # s = 2 * np.pi * freqs * 1j
        # resp = self.gain * self.norm_factor * np.ones(s.shape, np.complex128)
        # for p in self.poles:
        #     resp /= (s - p)
        # for z in self.zeros:
        #     resp *= (s - z)
        return resp

        s = 2 * np.pi * freqs * 1j
        hp = np.ones(s.shape, np.complex128)
        for p in self.poles:
            hp /= (s - p)
        for z in self.zeros:
            hp *= (s - z)
        return hp

    def plot(self, min_freq, output='VEL', show=True, label='', axes=None,
             sampling_rate=None, sym='b-'):
        """
        Plot PAZ object's Response

        :type min_freq: float
        :param min_freq: Lowest frequency to plot.
        :type output: str
        :param output: Output units. One of:

                ``"DISP"``
                    displacement
                ``"VEL"``
                    velocity
                ``"ACC"``
                    acceleration
        :type axes: list of 2 :class:`matplotlib.axes.Axes`
        :param axes: List/tuple of two axes instances to plot the
            amplitude/phase spectrum into. If not specified, a new figure is
            opened.
        :type label: str
        :param label: Label string for legend.
        :type sampling_rate: float
        :param sampling_rate: Manually specify sampling rate of time series.
            If not given it is attempted to determine it from the information
            in the individual response stages.  Does not influence the spectra
            calculation, if it is not known, just provide the highest frequency
            that should be plotted times two.
        :type show: bool
        :param show: Whether to show the figure after plotting or not. Can be
            used to do further customization of the plot before showing it.
        :type sym: str
        :param sym: symbol to use on plot
        """
        paz = self.copy()
        if output == 'DISP':
            paz.input_units = 'm'
        elif output == 'VEL':
            paz.input_units = 'm/s'
        elif output == 'ACC':
            paz.input_units = 'm/s^2'
        else:
            print(f'Unknown output type: {output}, using VEL')
            paz.input_units = 'm/s'
        if sampling_rate is None:
            sampling_rate = paz.sampling_rate
        if sampling_rate is None:
            if min_freq < 1:
                sampling_rate = 100.
            else:
                sampling_rate = min_freq * 100.
        else:
            assert sampling_rate > 2*min_freq
        if axes is not None:
            axs = axes
            fig = axs[0].figure
            npts = 50
        else:
            fig, axs = plt.subplots(2, 1, sharex=True)
            npts = 100
        f = np.power(10., np.linspace(np.log10(min_freq),
                                      np.log10(sampling_rate/2), npts))
        resp = paz.get_response(f)
        axs[0].loglog(f, np.absolute(resp), sym, label=label)
        axs[0].set_ylabel('Amplitude ({}/{})'.format(self.output_units.lower(),
                                                     self.input_units.lower()))
        axs[0].grid(True)
        axs[0].legend()
        axs[1].semilogx(f, np.angle(resp), sym)
        axs[1].grid(True)
        axs[1].set_ylabel('Phase')
        axs[1].yaxis.set_major_locator(plt.MultipleLocator(np.pi / 2))
        axs[1].yaxis.set_major_formatter(plt.FuncFormatter(
            multiple_formatter()))
        axs[1].set_xlabel('Frequency (Hz)')
        if show is True:
            plt.show()

    def calc_norm_factor(self):
        """
        Calculate A0 normalization factor

        Returns A0 such that A0 * prod(s-z_n)/prod(s-p_n) = 1 at ref_freq
        (SEED recommendation)
        """
        hp = self._hp([self.ref_freq])
        val = 1./abs(hp[0])
        # s = 2 * np.pi * self.ref_freq * 1j
        # val = 1
        # for p in self.poles:
        #     val *= (s - p)
        # for z in self.zeros:
        #     val /= (s - z)
        # val = abs(val)
        return val

    def to_obspy(self):
        """
        Return obspy paz dict

        obspy is not clear about what is in the "gain" and "sensitivity"
        keys.  I tried self.norm_factor and self.gain, respectively, but
        this makes their product depend on f_ref.  Switched to:
            self.norm_factor, self.ref_gain
        """
        return {'poles': list(self.poles), 'zeros': list(self.zeros),
                'f_ref': self.ref_freq, 'gain': self.norm_factor,
                'sensitivity': self.ref_gain}

    def _unit_conversion(self, new_input_units):
        """
        Convert from self.input_units to new_input_units
        """
        igain, inz = self._ref_meter(self.input_units)
        ogain, onz = self._ref_meter(new_input_units)

        # print(f'{self.input_units=}, {igain=}, {inz=}')
        # print(f'{new_input_units=}, {ogain=}, {onz=}')

        if igain is None or ogain is None:
            warnings.warn('Did not convert')
            return

        self.gain *= ogain / igain
        # if self._norm_factor is not None:
        #     old_calc_norm_factor = self.calc_norm_factor()
        relnz = onz - inz
        if relnz > 0:
            self.zeros = np.append(self.zeros, np.zeros(relnz))
        elif relnz < 0:
            relnp = -relnz
            i_zero_zeros = np.where(np.absolute(self.zeros) == 0)[0]
            if len(i_zero_zeros) > 0:
                if len(i_zero_zeros) >= relnp:
                    self.zeros = np.delete(self.zeros, i_zero_zeros[:relnp])
                    relnp = 0
                else:
                    self.zeros = np.delete(self.zeros, i_zero_zeros)
                    relnp -= len(i_zero_zeros)
            self.poles = np.append(self.poles, np.zeros(relnp))
        # if self._norm_factor is not None:
        #     # Since norm factor cannot change, change gain
        #     self.gain *= old_calc_norm_factor / self.calc_norm_factor()

    @classmethod
    def from_obspy_PoleZeroResponseStage(cls, stage):
        """
        Return PAZ correpsonding to obspy PoleZeroResponseStage
        """
        assert isinstance(stage, PolesZerosResponseStage)
        assert stage.stage_gain_frequency == stage.normalization_frequency

        sampling_rate = None
        if stage.decimation_input_sample_rate is not None:
            sampling_rate = stage.decimation_input_sample_rate
            if stage.decimation_factor is not None:
                sampling_rate /= stage.decimation_factor

        cls = PAZ(poles=stage.poles, zeros=stage.zeros,
                  input_units=stage.input_units,
                  output_units=stage.output_units,
                  ref_freq=stage.stage_gain_frequency,
                  gain=stage.stage_gain,
                  norm_factor=stage.normalization_factor,
                  sampling_rate=sampling_rate)
        return cls

    @classmethod
    def from_obspy_response(cls, resp):
        """
        Return PAZ fitting an obspy response object

        Combines all poles and zeros, multiplies all gains
        Does not fit frequency responses of other stage types
        """
        # print(f'{resp=}')
        obspy_paz = _get_pzrs_all(resp)
        # print(f'{obspy_paz=}')
        cls = PAZ.from_obspy_PoleZeroResponseStage(obspy_paz)
        return cls

    def write_json_pz(self, filename):
        """
        write JSON paz format
        """
        if not self.output_units.lower() == 'counts':
            print("Can't write to json_pz: output_units are "
                  f"{self.output_units}, not counts")
            return
        out_paz = self.copy()
        out_paz.input_units = 'm/s'
        # out_paz.ref_freq=.1
        outdict = {'poles': [[x.real, x.imag] for x in out_paz.poles],
                   'zeros': [[x.real, x.imag] for x in out_paz.zeros],
                   'sensitivity': 1/out_paz.ref_gain,
                   'input_units': out_paz.input_units,
                   'norm_const': out_paz.norm_factor,
                   'f_ref': out_paz.ref_freq}
        if out_paz.sampling_rate is not None:
            outdict['sampling_rate'] = out_paz.sampling_rate
        with open(filename, 'w') as f:
            try:
                json.dump(outdict, f, indent=2)
            except Exception as e:
                raise ValueError(f'Could not write to {filename}: {e}')

    @classmethod
    def read_json_pz(cls, filename):
        """
        read JSON paz format
        """
        with open(filename) as f:
            try:
                paz = json.load(f)
            except Exception as e:
                raise ValueError(f'Could not load {filename}: {e}')
        poles, zeros = [], []
        for p in paz['poles']:
            poles.append(float(p[0]) + float(p[1])*1.j)
        for z in paz['zeros']:
            zeros.append(float(z[0]) + float(z[1])*1.j)
        val = cls(poles=poles, zeros=zeros,
                  ref_gain=1./float(paz['sensitivity']),
                  ref_freq=float(paz['f_ref']),
                  input_units=paz.get('input_units', 'm/s'),
                  output_units='counts',
                  norm_factor=paz.get('norm_const', None),
                  sampling_rate=paz.get('sampling_rate', None))
        return val

    @classmethod
    def read_stationxml(cls, filename, channel, station='*'):
        """
        read JSON paz format
        """
        inv = read_inventory(filename, 'STATIONXML')
        inv = inv.select(channel=channel, station=station)
        nscs = [{'net': n.code, 'sta': s.code, 'chan': c}
                for n in inv for s in n for c in s]
        if len(nscs) > 1:
            nsc_strs = [f'{x["net"]}.{x["sta"]}.{x["chan"].code}'
                        for x in nscs]
            print('Multiple channels were selected, using {nsc_strs[0]}')
            print('Other channels were: ' + ','.join(nsc_strs[1:]))
        resp = nscs[0]["chan"].response
        return cls.from_obspy_response(resp)

    @classmethod
    def read_sac_pz(cls, filename):
        """
        read SAC PZ format

        input units are assumed to be meters (does not look at comments)
        The file-reading code is copied from obspy.io.sac.sacpz.attach_paz()
        """
        with open(filename) as f:
            poles = []
            zeros = []
            while True:
                line = f.readline()
                if not line:
                    break
                # lines starting with * are comments
                if line.startswith('*'):
                    continue
                if line.find('ZEROS') != -1:
                    a = line.split()
                    noz = int(a[1])
                    for _k in range(noz):
                        line = f.readline()
                        a = line.split()
                        if line.find('POLES') != -1 or \
                           line.find('CONSTANT') != -1 or \
                           line.startswith('*') or not line:
                            while len(zeros) < noz:
                                zeros.append(complex(0, 0j))
                            break
                        else:
                            zeros.append(complex(float(a[0]), float(a[1])))

                if line.find('POLES') != -1:
                    a = line.split()
                    nop = int(a[1])
                    for _k in range(nop):
                        line = f.readline()
                        a = line.split()
                        if line.find('CONSTANT') != -1 or \
                           line.find('ZEROS') != -1 or \
                           line.startswith('*') or not line:
                            while len(poles) < nop:
                                poles.append(complex(0, 0j))
                            break
                        else:
                            poles.append(complex(float(a[0]), float(a[1])))
                if line.find('CONSTANT') != -1:
                    a = line.split()
                    constant = float(a[1])
        # constant is norm_factor * gain, so if I set one to constant, I have
        # to set the other to 1.
        val = cls(poles=poles,
                  zeros=zeros,
                  gain=constant,
                  norm_factor=1,
                  input_units='m',
                  output_units='counts')
        return val

    @classmethod
    def read_gse_response(cls, filename):
        """
        read a GSE response file

       From specification in GSE_2.0.pdf (Group of Scientific Expernts, 1995)
        :param name:    Name of GSE2 response file
        :returns: Structure with all parameters
        """
        # read_GSE.m:6
        info = dict()
        with open(filename, 'r') as fic:
            for line in fic:
                # CAL2 line is specified in Table 13, page 69
                if line[:4] == 'CAL2':
                    info['station'] = line[5:10]
                    info['type'] = line[20:26].strip()
                    info['sensitivity'] = float(line[27:37])
                    info['f_ref'] = 1. / float(line[38:45])
                    info['sampling_rate'] = float(line[46:56])
                # PAZ2 line is specified in Table 14, page 70
                elif line[:4] == 'PAZ2':
                    info['norm_const'] = float(line[10:25])
                    num_poles = float(line[40:43])
                    num_zeros = float(line[44:47])
                    info['poles'] = []
                    info['zeros'] = []
                    for p in np.arange(num_poles):
                        A = next(fic).split()
                        info['poles'].append(np.complex(float(A[0]),
                                                        float(A[1])))
                    for z in np.arange(num_zeros):
                        A = next(fic).split()
                        info['zeros'].append(np.complex(float(A[0]),
                                                        float(A[1])))
                # DIG2 line is specified in Table 17, page 72
                elif line[:4] == 'DIG2':
                    info['amplifier_gain'] = float(line[8:23])
                    info['digitizer'] = line[36:].strip()
                else:
                    continue
        info['f_ref'] = info.get('f_ref', 1.)
        val = cls(poles=info['poles'], zeros=info['zeros'],
                  ref_gain=1./info['sensitivity'],
                  ref_freq=info['f_ref'],
                  input_units='nm',
                  output_units='counts',
                  norm_factor=info.get('norm_const', None),
                  sampling_rate=info['sampling_rate'])
        return val

    @classmethod
    def read_baillard_pz(cls, filename):
        """
        read Baillard PZ format

        A numbers-only version of the GSE format
        """
        info = {}
        with open(filename, 'r') as foc:
            i = 0
            a, info['poles'], info['zeros'] = [], [], []
            for tline in foc:
                if i < 4:
                    a.append(float(tline))
                elif i == 4:
                    n_poles = float(tline)
                elif i == 5:
                    n_zeros = float(tline)
                elif i >= 6 and i < (6 + n_poles):
                    A = tline.split()
                    info['poles'].append(np.complex(float(A[0]), float(A[1])))
                elif i >= 6 + n_poles and i < 6 + n_poles + n_zeros:
                    A = tline.split()
                    info['zeros'].append(np.complex(float(A[0]), float(A[1])))
                else:
                    info['amplifier_gain'] = float(tline)
                i += 1
        info['sensitivity'] = a[0]
        info['f_ref'] = 1 / a[1]
        info['sampling_rate'] = a[2]
        info['A0'] = a[3]
        val = cls(poles=info['poles'],
                  zeros=info['zeros'],
                  ref_gain=1/info['sensitivity'],
                  ref_freq=info['f_ref'],
                  input_units='nm',
                  output_units='counts',
                  norm_factor=info['A0'],
                  sampling_rate=info['sampling_rate'])
        return val

    def known_units(self):
        """
        Return known input_units (for automatic conversions
        """
        return list(self.converter.keys())

    def _ref_meter(self, units):
        """
        Return conversions needed to convert to meters

        :returns: gain_factor, number of zeros
        """
        if units.lower() in self.converter:
            return (self.converter[units.lower()][0],
                    self.converter[units.lower()][1])
        else:
            warnings.warn('Unknown conversion unit: {}, known values are "{}"'
                          .format(units, ', '.join(self.converter.keys())))
            return None, None


def _get_pzrs(resp, verbose=False):
    """
    Get PolesZerosResponseStages

    Like obspy get_paz(), except that. if there is more than one PZ stage,
    combines them into one

    Does not handle decimation, resource_id, resource_id2
    Returns LAST stage's stage_sequence number
    :param resp: obspy Response object
    :returns: obspy PoleZeroResponse stage (stage_sequence_number = LAST
        paz stage)
    """

    pz_stages = [copy.deepcopy(stage) for stage in resp.response_stages
                 if isinstance(stage, PolesZerosResponseStage)]
    if len(pz_stages) == 0:
        print("No PolesZerosResponseStage found.")
        pz_out = PolesZerosResponseStage(normalization_factor=1.,
                                         stage_gain=1.,
                                         poles=[],
                                         zeros=[])
    elif len(pz_stages) == 1:
        pz_out = pz_stages[0]
    else:
        if verbose:
            print(f"Combining {len(pz_out):d} PolesZerosResponseStages")
        pz_out = pz_stages[0]
        for pz in pz_stages[1:]:
            if _is_compatible(pz_out, pz):
                if pz.description is not None:
                    pz_out.description += ' + ' + pz.description
                if not pz_out.name:
                    if pz.name:
                        pz_out.name = ' + ' + pz.name
                elif pz.name:
                    pz_out.name += ' + ' + pz.name
                pz_out.normalization_factor *= pz.normalization_factor
                pz_out.output_units = pz.output_units
                pz_out.output_units_description = pz.output_units_description
                pz_out.stage_gain *= pz.stage_gain
                pz_out.poles.extend(pz.poles)
                pz_out.zeros.extend(pz.zeros)
                pz_out.stage_sequence_number = pz.stage_sequence_number
            else:
                print('Incompatible stages... quitting')
                return None
    return pz_out


def _get_pzrs_all(resp):
    """
    Get PoleZeroResponseStage covering ALL stages

    :param resp: obspy Response object
    :returns: obspy PoleZeroResponse stage (stage_sequence_number = LAST
        PoleZeroResponseStage)
    """
    pz_out = _get_pzrs(resp)
    for stage in resp.response_stages[pz_out.stage_sequence_number:]:
        if _is_compatible(pz_out, stage, next_is_PAZ=False):
            pz_out.output_units = stage.output_units
            pz_out.output_units_description = stage.output_units_description
            pz_out.stage_gain *= stage.stage_gain
            pz_out.stage_sequence_number = stage.stage_sequence_number
            pz_out.decimation_input_sample_rate =\
                stage.decimation_input_sample_rate
            pz_out.decimation_factor = stage.decimation_factor
        else:
            print('Incompatible stages... quitting')
            return None
    if pz_out.description is None:
        pz_out.description = 'PAZ stages'
    pz_out.description += ' + non-PAZ stages'
    if pz_out.name:
        pz_out.name += ' + non-PAZ stages'
    return pz_out


def _is_compatible(first_stage, next_stage, next_is_PAZ=True):
    """Can one PAZ stage be safely added to another?"""
    if not first_stage.output_units == next_stage.input_units:
        msg = ("First stage out units not equal to next stage in units. ")
        warnings.warn(msg)
        return False
    if next_is_PAZ:
        if not (first_stage.pz_transfer_function_type
                == next_stage.pz_transfer_function_type):
            msg = ("Stage transfer function types not equal")
            warnings.warn(msg)
            return False
        if not (first_stage.stage_gain_frequency
                == next_stage.stage_gain_frequency):
            print("Stage gain frequencies not equal ({:g} and {:g})".format(
                first_stage.stage_gain_frequency,
                next_stage.stage_gain_frequency))
            return False
        if not (first_stage.normalization_frequency
                == next_stage.normalization_frequency):
            msg = ("Normalization frequencies not equal")
            warnings.warn(msg)
            return False
        if first_stage.decimation_factor or next_stage.decimation_factor:
            msg = ("Can't handle decimation (yet)")
            warnings.warn(msg)
            return False
    return True


def multiple_formatter(denominator=2, number=np.pi, latex='\pi'):
    def gcd(a, b):
        while b:
            a, b = b, a % b
        return a

    def _multiple_formatter(x, pos):
        den = denominator
        num = np.int(np.rint(den*x/number))
        com = gcd(num, den)
        (num, den) = (int(num/com), int(den/com))
        if den == 1:
            if num == 0:
                return r'$0$'
            if num == 1:
                return r'$%s$' % latex
            elif num == -1:
                return r'$-%s$' % latex
            else:
                return r'$%s%s$' % (num, latex)
        else:
            if num == 1:
                return r'$\frac{%s}{%s}$' % (latex, den)
            elif num == -1:
                return r'$\frac{-%s}{%s}$' % (latex, den)
            else:
                return r'$\frac{%s%s}{%s}$' % (num, latex, den)
    return _multiple_formatter



# Generated with SMOP  0.41-beta
# from smop.libsmop import *
# readmain.m


class PolarityParameters():
    """
    Polarity Parameters
    """
    def __init__(self,
                 calculate_window=2,
                 analyze_window=4,
                 DR_threshold_P=0.4,
                 DR_threshold_S=-0.4,
                 DR_smooth_length=1):
        """
        Initialize Polarity Parameters

        :param calculate_window: number of seconds AFTER a pick in which
            to calculate polarities
        :param analyze_window: number of seconds to analyze for each
            polarity calculation
        :param DR_threshold_P: DipRect P-wave threshold
        :param DR_threshold_S: DipRect S-wave threshold
        :param DR_smooth_length: seconds over which to smooth polarity and
            rectilinearity when calculating dip-rect
        """
        self.calculate_window = calculate_window
        self.analyze_window = analyze_window
        self.DR_threshold_P = DR_threshold_P
        self.DR_threshold_S = DR_threshold_S
        self.DR_smooth_length = DR_smooth_length

    def __str__(self):
        str = "PolarityParameters:\n"
        str += f"    calculate_window = {self.calculate_window}\n"
        str += f"    analyze_window = {self.analyze_window}\n"
        str += f"    DR_threshold_P = {self.DR_threshold_P}\n"
        str += f"    DR_threshold_S = {self.DR_threshold_S}\n"
        str += f"    DR_smooth_length = {self.DR_smooth_length}\n"
        return str

    @classmethod
    def from_dict(cls, thedict):
        return(cls(**thedict))


if __name__ == '__main__':
    pass
